# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

from pathlib import Path


def format_npz_samples_filename(start_id: int, num_samples: int) -> str:
    """Format the filename for the samples npz file."""
    # no good way to find a width that works for all cases. Just guessing that 1000000 should be large enough.
    return f"batch_{start_id:07d}_{start_id + num_samples:07d}.npz"


def count_samples_in_output_dir(output_dir: Path) -> int:
    """Count the number of samples in the npz files in the output directory.
    assumption: the samples files are named as batch_<start_id>_<end_id+1>.npz
    """
    num_samples = [
        int(pair[1]) - int(pair[0])
        for p in output_dir.glob("batch_*.npz")
        for pair in [p.stem.split("_")[1:]]
    ]
    return sum(num_samples)



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
"""
This is an integration test of the denoising algorithm. For a known data distribution that
is Gaussian, we sample using the analytically derived ground truth score and check
we retrieve correct moments of the data distribution.
"""


import pytest
import torch
from torch_geometric.data import Batch

from bioemu.chemgraph import ChemGraph
from bioemu.denoiser import dpm_solver, heun_denoiser
from bioemu.sde_lib import CosineVPSDE
from bioemu.so3_sde import DiGSO3SDE, rotmat_to_rotvec


@pytest.mark.parametrize(
    "solver,denoiser_kwargs", [(dpm_solver, {}), (heun_denoiser, {"noise": 0.5})]
)
def test_reverse_sampling(solver, denoiser_kwargs):
    torch.manual_seed(1)
    N = 200  # Timesteps
    batch_size = 1000

    # Assume a ground truth distribution for 1D positions to be of mean -3 and std 4.3
    x0_mean = torch.tensor(-3.0)
    x0_std = torch.tensor(4.3)

    r3sde = CosineVPSDE()
    so3sde = DiGSO3SDE(num_sigma=10)
    sdes = {"pos": r3sde, "node_orientations": so3sde}

    def node_orientation_score(Rt: torch.Tensor, t: torch.Tensor):
        # Assume a ground truth distritubution for SO(3) is a delta distribution at the identity.
        # Note we use the gradient rotation vector for score instead of the matrix.
        grad_coefficients = so3sde.compute_score(rotmat_to_rotvec(Rt), t)
        assert grad_coefficients.shape == (batch_size, 3)
        return grad_coefficients

    def pos_score(x_t: torch.Tensor, t: torch.Tensor) -> torch.Tensor:
        a_t, s_t = r3sde.marginal_prob(x=torch.ones_like(x_t), t=t)
        x0 = (x0_mean * s_t**2 + x_t * a_t * x0_std**2) / (s_t**2 + a_t**2 * x0_std**2)
        return (x0 * a_t - x_t) / s_t

    def score_fn(x: ChemGraph, t: torch.Tensor) -> ChemGraph:
        expected_x0 = {
            "pos": pos_score(x.pos, t),
            "node_orientations": node_orientation_score(x.node_orientations, t),
        }
        return x.replace(**expected_x0)

    conditioning_data = Batch.from_data_list(
        [
            ChemGraph(
                pos=torch.randn(batch_size, 1),
                node_orientations=so3sde.prior_sampling((batch_size,)),
            )
        ]
    )

    samples = solver(
        sdes=sdes,
        batch=conditioning_data,
        N=N,
        score_model=score_fn,
        max_t=0.99,
        eps_t=0.001,
        device=torch.device("cpu"),
        **denoiser_kwargs,
    )

    assert torch.isclose(samples.pos.mean(), x0_mean, rtol=1e-1, atol=1e-1)
    assert torch.isclose(samples.pos.std().mean(), x0_std, rtol=1e-1, atol=1e-1)
    assert torch.allclose(samples.node_orientations.mean(dim=0), torch.eye(3), atol=1e-1)
    assert torch.allclose(samples.node_orientations.std(dim=0), torch.zeros(3, 3), atol=1e-1)



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
# Quick way to refer to things to instantiate in the config
from .denoiser import dpm_solver, heun_denoiser  # noqa
from .models import DiGConditionalScoreModel  # noqa
from .sde_lib import CosineVPSDE  # noqa
from .so3_sde import DiGSO3SDE  # noqa



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

import math

import numpy as np
import torch
from torch import nn
from torch_geometric.utils import to_dense_adj, to_dense_batch

from .chemgraph import ChemGraph
from .structure_module import StructureModule

# Number of dimensions for the nodes and edges of Evoformer embeddings
EVOFORMER_NODE_DIM: int = 384
EVOFORMER_EDGE_DIM: int = 128


class SinusoidalPositionEmbedder(nn.Module):
    def __init__(
        self,
        dim: int,
        max_period: int = 10000,
        min_input: float = 0.0,
        max_input: float = 1000.0,  # 1000 is the DiG default maximum diffusion time value.
    ) -> None:
        """
        Sinusoidal embedding for encoding a scalar input.
        See e.g.
        https://pytorch.org/tutorials/beginner/transformer_tutorial.html

        Args:
            dim: Dimension of embedding.
            max_period: Maximum period of embedding.
            min_input: The minimum expected range of the scalar input variable.
            max_input: The maximum expected range of the scalar input variable.
        """
        super().__init__()
        self.dim = dim
        self.half_dim = self.dim // 2
        self.min_input = min_input
        self.max_input = max_input

        self.embedding_factor = -math.log(max_period) / (self.half_dim - 1)

        self.dummy = nn.Parameter(
            torch.empty(0, dtype=torch.float), requires_grad=False
        )  # to detect fp16

    def forward(self, time: torch.Tensor) -> torch.Tensor:
        """
        Convert diffusion times to the corresponding embeddings.

        Args:
            time: Tensor of times.

        Returns:
            Time embeddings.
        """
        # Rescale the input scalar so that it is in the range [0, 1000.]. This matches the
        # behaviour of the timestep embedder in DiG.
        time = (time - self.min_input) * 1000.0 / (self.max_input - self.min_input)

        device = time.device
        embeddings = torch.exp(torch.arange(self.half_dim, device=device) * self.embedding_factor)
        embeddings = time[:, None] * embeddings[None, :]
        embeddings = torch.cat((embeddings.sin(), embeddings.cos()), dim=-1)
        embeddings = embeddings.to(self.dummy.dtype)
        return embeddings


class RelativePositionBias(nn.Module):
    def __init__(self, num_buckets: int = 64, max_distance: int = 256, out_dim: int = 2) -> None:
        """
        Module for computing embeddings based on the relative position between residues in the
        sequence. Distances are categorized into different buckets and each is assigned a learnable
        embedding.

        NOTE: The algorithm here for computing the embeddings is distinct from both AF2 and FrameDiff. It is taken from DiG.
        The resolution of the buckets gets lower the
        further you go from the diagonal.

        Args:
            num_buckets: Number of buckets used for binning pair distances.
            max_distance: Maximum distance considered for embedding.
            out_dim: Dimension of output embeddings.
        """
        super().__init__()
        self.num_buckets = num_buckets
        self.max_distance = max_distance
        self.relative_attention_bias = nn.Embedding(self.num_buckets, out_dim)

    @staticmethod
    def _relative_position_bucket(
        relative_position: torch.Tensor, num_buckets: int, max_distance: int
    ) -> torch.Tensor:
        """
        Obtain embedding index based on current relative positions.

        Args:
            relative_position: Relative positions in sequence.
            num_buckets: Number of buckets used for binning.
            max_distance: Maximum distance considered for embedding.

        Returns:
            Index corresponding to the embedding vector of each pairwise distance.
        """
        num_buckets //= 2
        ret = (relative_position < 0).to(relative_position) * num_buckets
        relative_position = torch.abs(relative_position)
        max_exact = num_buckets // 2
        is_small = relative_position < max_exact

        val_if_large = (
            max_exact
            + (
                torch.log(relative_position / max_exact)
                / math.log(max_distance / max_exact)
                * (num_buckets - max_exact)
            ).long()
        )
        val_if_large = torch.min(val_if_large, torch.full_like(val_if_large, num_buckets - 1))

        ret += torch.where(is_small, relative_position, val_if_large)
        return ret

    def forward(self, relative_position: torch.Tensor) -> torch.Tensor:
        """
        Get relative position embeddings.

        Args:
            relative_position: Tensor of relative sequence positions.

        Returns:
            Embedding vectors corresponding to relative positions.
        """
        rp_bucket = self._relative_position_bucket(  # [L, L] integers
            relative_position,
            num_buckets=self.num_buckets,
            max_distance=self.max_distance,
        )
        rp_bias = self.relative_attention_bias(
            rp_bucket
        )  # [L, L, C] embeddings of each relative sequence distance
        return rp_bias


class DistributionalGraphormer(nn.Module):
    def __init__(
        self,
        dim_model: int = 512,
        dim_pair: int = 256,
        num_layers: int = 8,
        num_heads: int = 32,
        dim_single_rep: int = 64,
        dim_hidden: int = 1024,
        num_buckets: int = 64,
        max_distance_relative: int = 128,
        dropout: float = 0.1,
    ):
        """
        Basic distributional graphormer model. For architecture details, please refer to:
        https://arxiv.org/abs/2306.05445

        Args:
            dim_model: Number of features used for main model representation.
            dim_pair: Number of features used for pair representations.
            num_layers: Number of layers used in the structure model.
            num_heads: Number of heads used for attention.
            dim_single_rep: Number of features used for single sequence embeddings.
            dim_pair_rep: Number of features used for pair embeddings.
            dim_hidden: Number of nodes used in hidden layers.
            num_buckets: Number of buckets used in relative positional encoding.
            max_distance_relative: Maximum distance considered in relative positional encoding.
            dropout: _description_. Defaults to 0.1.

            Let (T_in, R_in) be the input frames, and let (T, R) be an arbitrary rotation and translation.
            Let T_out, R_out = model(T_in, R_in). And let T_out_transformed, R_out_transformed = model((T, R) * (T_in, R_in)), where frames are
            composed according to Section 1.1 of the AF2 Supplementary Material.

            The model has the following equivariance properties:
                T_out_transformed = R * T_out
                R_out_transformed = R_out
            This makes it suitable for regressing the "invariant score", as in the DiG paper.
        """

        super().__init__()

        self.d_model = dim_model

        # Set the dimensions for EVOFORMER feature embedding
        dim_single_rep = EVOFORMER_NODE_DIM
        self.dim_pair_rep = EVOFORMER_EDGE_DIM

        self.step_emb = SinusoidalPositionEmbedder(dim=self.d_model)
        self.x1d_proj = nn.Sequential(
            nn.LayerNorm(dim_single_rep),
            nn.Linear(dim_single_rep, self.d_model, bias=False),
        )
        self.x2d_proj = nn.Sequential(
            nn.LayerNorm(self.dim_pair_rep),
            nn.Linear(self.dim_pair_rep, dim_pair, bias=False),
        )
        self.rp_proj = RelativePositionBias(
            num_buckets=num_buckets, max_distance=max_distance_relative, out_dim=dim_pair
        )

        self.st_module = StructureModule(
            d_pair=dim_pair,
            n_layer=num_layers,
            d_model=self.d_model,
            n_head=num_heads,
            dim_feedforward=dim_hidden,
            dropout=dropout,
        )

    def forward(
        self,
        x: torch.Tensor,
        node_orientations: torch.Tensor,
        batch_index: torch.Tensor,
        t: torch.Tensor,
        context: ChemGraph,
    ) -> tuple[torch.Tensor, torch.Tensor]:
        """
        Predict translation and rotation scores based on noisy positions and node orientations.
        This function takes and returns quantities in sparse batch format.

        Args:
            x: [N, 3] node positions.
            node_orientations: Rotation matrices encoding orientations [N, 3, 3].
            batch_index: Sparse batch indices.
            t: Current diffusion time steps.
            context: Current batch in ChemGraph format. Used for embeddings, labels and masking.

        Returns:
            Predicted translation and rotation scores, both in vector representation. The
            translation score will be equivariant and the rotation score invariant to global pose
            rotations.
        """
        batch_index = context.batch
        # Convert from sparse graph representation to dense representation used by DiG.
        T_perturbed, mask = to_dense_batch(
            x, batch_index
        )  # [B, L, 3], [B, L], L is maximum number of residues in a batch
        IR_perturbed, _ = to_dense_batch(node_orientations, batch_index)  # [B, L, 3, 3]

        # Transform node and edge embeddings from sparse to dense.
        single_repr, _ = to_dense_batch(context.single_embeds, batch_index)
        single_repr = single_repr.to(torch.float32)  # [B, L, 384]
        pair_repr = to_dense_adj(  # [B, L, L, 128]
            context.edge_index, batch_index, edge_attr=context.pair_embeds
        ).to(torch.float32)

        # Deal with masking that comes from the data itself, not just the dense batch representation.
        # There are two sources of masking - the first is the mask that comes from the data itself,
        # which represents residues where the position/orientation are unknown. This is captured in
        # `pos_is_known`. The second is masking due to uneven length proteins in the dense batch
        # representation. This is captured in `mask``.
        # attn_mask is True if the element should be masked out in attention, following DiG conventions.
        if "pos_is_known" in context:
            pos_is_known = context.pos_is_known
            pos_is_known_dense = to_dense_batch(pos_is_known, batch_index)[0].bool()  # [B, L]
            attn_mask = ~(mask & pos_is_known_dense)  # [B, L]
        else:
            attn_mask = ~mask  # [B, L]

        t, _ = to_dense_batch(t, batch_index)  # [B, L]
        t = t[:, 0]  # [B], each residue experiences the same timestep.

        # Embed single_repr, noise level "t" and the conditionings from context into x1d.
        x1d = self.x1d_proj(single_repr) + self.step_emb(t)[:, None]  # [B, L, C]

        x2d = self.x2d_proj(pair_repr)  # [B, L, L, C]

        pos_sequence = torch.arange(
            T_perturbed.shape[1], device=x1d.device
        )  # [L], integer sequence index

        pos_sequence = pos_sequence.unsqueeze(1) - pos_sequence.unsqueeze(
            0
        )  # [L, L], integer relative sequence positions.
        x2d = x2d + self.rp_proj(pos_sequence)[None]

        z = (~attn_mask).long().sum(-1, keepdims=True)  # [B, 1], number of unmasked elements.
        filled_mask = attn_mask.masked_fill(
            z == 0, False  # z == 0 means no unmasked elements, i.e., all elements are masked.
        )  # [B, L], fill in False if all elements are masked. This is to avoid the case where all elements are masked, which would cause an error in the softmax I think.

        bias = filled_mask.float().masked_fill(filled_mask, float("-inf"))[:, None, :, None]
        bias = bias.permute(
            0, 3, 1, 2
        )  # [B, 1, 1, L], this has the value -inf for elements that are masked.

        # Change in translation and rotation. At this point, both quantities are invariant to global
        # SE(3) transformations.
        (
            T_eps,
            IR_eps,
        ) = self.st_module(  # st_module plays an equivalent role to BackboneUpdate in the Algorithm 20 of AF2 supplement.
            (T_perturbed, IR_perturbed), x1d, x2d, bias
        )

        # Introduce orientation dependence of the translation score.
        T_eps = torch.matmul(IR_perturbed.transpose(-1, -2), T_eps.unsqueeze(-1)).squeeze(-1)
        T_out, R_out = T_eps, IR_eps

        # Return back to sparse graph representation
        T_out = T_out[mask]  # [N, 3]
        R_out = R_out[mask]  # [N, 3], in axis angle representation.

        return (
            T_out,
            R_out,
        )

    def __str__(self) -> str:
        """
        Model prints with number of trainable parameters.
        """
        model_parameters = filter(lambda p: p.requires_grad, self.parameters())
        params = sum([np.prod(p.size()) for p in model_parameters])
        return super().__str__() + f"\nTrainable parameters: {params}"


class DiGConditionalScoreModel(torch.nn.Module):
    """Wrapper to convert the DiG nn.Module neural network that operates directly on position
    and rotation tensors into a ScoreModel that operates on ChemGraph objects.
    """

    def __init__(
        self,
        dim_model: int = 512,
        dim_pair: int = 256,
        num_layers: int = 8,
        num_heads: int = 32,
        dim_single_rep: int = 64,
        dim_hidden: int = 1024,
        num_buckets: int = 64,
        max_distance_relative: int = 128,
        dropout: float = 0.1,
    ):
        """
        Args: all passed through to DistributionalGraphormer
        """
        super().__init__()
        self.model_nn = DistributionalGraphormer(
            dim_model=dim_model,
            dim_pair=dim_pair,
            num_layers=num_layers,
            num_heads=num_heads,
            dim_single_rep=dim_single_rep,
            dim_hidden=dim_hidden,
            num_buckets=num_buckets,
            max_distance_relative=max_distance_relative,
            dropout=dropout,
        )

    def forward(self, x: ChemGraph, t: torch.Tensor) -> ChemGraph:
        # NOTE: the DiG structure model uses a time embedding intended for integer time
        # steps between 0 and num_timesteps (1000 by default). The SDE gets times between
        # 0 and 1, where the filter used in the embedding is less expressive. To avoid
        # this, t is scaled by the timesteps before passing to model.
        assert hasattr(x, "batch"), "batch of ChemGraphs must have a 'batch' attribute."
        time_effective = t[x.batch] * 1000
        # NOTE: DiG takes in inverse rotations as input. To be consistent
        # with frame conventions and sampling in the rest of the code, frames are transposed
        # / inverted here.
        node_orientations_effective = x.node_orientations.swapaxes(-1, -2)

        context = x.replace(pos=None, node_orientations=None)

        # pos is the translation score, node_orientations is the rotation score in axis
        # angle representation.
        pos_effective = x.pos
        (pos, node_orientations) = self.model_nn(
            x=pos_effective,
            node_orientations=node_orientations_effective,
            batch_index=x.batch,
            t=time_effective,
            context=context,
        )

        return x.replace(pos=pos, node_orientations=node_orientations)



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
import numpy as np
import pytest
import torch
from scipy.integrate import cumulative_trapezoid
from scipy.spatial.transform import Rotation

from bioemu.so3_sde import (
    SampleIGSO3,
    SampleUSO3,
    ScoreSO3,
    _broadcast_identity,
    _rotquat_to_axis_angle,
    angle_from_rotmat,
    apply_rotvec_to_rotmat,
    digso3_expansion,
    dlog_igso3_expansion,
    igso3_expansion,
    integrate_trapezoid_cumulative,
    rotmat_to_rotvec,
    rotquat_to_rotmat,
    rotquat_to_rotvec,
    rotvec_to_rotmat,
    scale_rotmat,
    skew_matrix_exponential_map,
    skew_matrix_exponential_map_axis_angle,
    skew_matrix_to_vector,
    vector_to_skew_matrix,
)

NUM_ROTATIONS = 10
NUM_OMEGA = 500
NUM_SIGMA = 500
NUM_SAMPLES = 10


@pytest.fixture
def rotation_angles(num_rotations=NUM_ROTATIONS):
    """Generate rotation angles in the interval [5e-2, pi-1e-2]."""
    angles = np.random.uniform(5e-2, np.pi - 1e-2, size=(num_rotations,))
    angles = torch.from_numpy(angles)
    return angles


@pytest.fixture
def rotation_vectors(rotation_angles):
    """
    Create rotation vectors by uniformly sampling rotation axes and scaling their length to the
    target rotation angles.
    """
    vectors = np.random.normal(size=(rotation_angles.shape[0], 3))
    vectors = vectors / np.linalg.norm(vectors, axis=1, keepdims=True)
    vectors = torch.from_numpy(vectors)
    rotvecs = rotation_angles[:, None] * vectors
    return rotvecs


@pytest.fixture
def rotation_matrices(rotation_vectors):
    """Convert rotation vectors to matrices using Scipy."""
    rotation = Rotation.from_rotvec(rotation_vectors)
    rotmats = rotation.as_matrix()
    rotmats = torch.from_numpy(rotmats)
    return rotmats


@pytest.fixture
def rotation_quaternions(rotation_vectors):
    """Convert rotation vectors to quaternions using Scipy."""
    rotation = Rotation.from_rotvec(rotation_vectors)
    rotquats = rotation.as_quat()
    rotquats = torch.from_numpy(rotquats)
    # Scipy uses the convention [i,j,k,r] for quaternions, shift to [r,i,j,k].
    rotquats = torch.roll(rotquats, 1, dims=-1)
    return rotquats


@pytest.fixture
def skew_matrices(rotation_vectors):
    """Convert rotation vectors to skew matrix form."""
    skewmats = torch.zeros(*rotation_vectors.shape, 3, dtype=rotation_vectors.dtype)
    skewmats[..., 2, 1] = rotation_vectors[..., 0]
    skewmats[..., 0, 2] = rotation_vectors[..., 1]
    skewmats[..., 1, 0] = rotation_vectors[..., 2]
    skewmats = skewmats - skewmats.transpose(-2, -1)
    return skewmats


def test_integrate_trapezoid_cumulative() -> None:
    """Test for cumulative distribution function integration."""
    # Generate unequally spaced grid from 0 to 10.
    x = torch.linspace(0.0, 1.0, 100) ** 3
    x = x[None, :] * 10.0

    # Compute sinc function on grid.
    y = torch.sinc(x)

    int_y = integrate_trapezoid_cumulative(y, x)

    # Get Scipy reference and bring to torch conventions.
    int_y_ref = cumulative_trapezoid(y[0].numpy(), x[0].numpy())
    int_y_ref = torch.from_numpy(int_y_ref)[None, :]

    assert torch.allclose(int_y, int_y_ref)


def test_broadcast_identity(rotation_matrices: torch.Tensor) -> None:
    """Test identity matrix broadcasting."""
    id3 = _broadcast_identity(rotation_matrices)

    # Check shape.
    assert id3.shape == rotation_matrices.shape

    # Check diagonals are 1.
    diag = torch.diagonal(id3, dim1=-2, dim2=-1)
    assert torch.allclose(
        diag, torch.tensor(1.0, dtype=rotation_matrices.dtype, device=rotation_matrices.device)
    )

    # Check offdiagonals are 0
    offdiag = id3[..., torch.eye(3) != 1]
    assert torch.allclose(
        offdiag, torch.tensor(0.0, dtype=rotation_matrices.dtype, device=rotation_matrices.device)
    )


def test_vector_to_skew_matrix(rotation_vectors: torch.Tensor, skew_matrices: torch.Tensor) -> None:
    """Test construction of skew matrices from vectors."""
    skewmats = vector_to_skew_matrix(rotation_vectors)
    assert torch.allclose(skewmats, skew_matrices)


def test_skew_matrix_to_vector(rotation_vectors: torch.Tensor, skew_matrices: torch.Tensor) -> None:
    """Test extraction of vectors from skew matrices."""
    rotvecs = skew_matrix_to_vector(skew_matrices)
    assert torch.allclose(rotvecs, rotation_vectors)


def test_skew_matrix_exponential_map_axis_angle(
    rotation_vectors: torch.Tensor, rotation_matrices: torch.Tensor
) -> None:
    """Test skew matrix exponential computation."""
    angles = torch.norm(rotation_vectors, dim=-1)
    vectors = rotation_vectors / angles[..., None]

    # Exponentials using Rodrigues' formula.
    skewmats = vector_to_skew_matrix(vectors)
    exp_map = skew_matrix_exponential_map_axis_angle(angles, skewmats)

    # Reference matrix exponential
    exp_ref = torch.linalg.matrix_exp(angles[..., None, None] * skewmats)

    assert torch.allclose(exp_map, exp_ref)
    assert torch.allclose(exp_map, rotation_matrices)


def test_skew_matrix_exponential_map(
    rotation_vectors: torch.Tensor, rotation_matrices: torch.Tensor
) -> None:
    """Test skew matrix exponential computation."""
    angles = torch.norm(rotation_vectors, dim=-1)

    # Exponentials using Rodrigues' formula.
    skewmats = vector_to_skew_matrix(rotation_vectors)
    exp_map = skew_matrix_exponential_map(angles, skewmats)

    # Reference matrix exponential
    exp_ref = torch.linalg.matrix_exp(skewmats)

    assert torch.allclose(exp_map, exp_ref)
    assert torch.allclose(exp_map, rotation_matrices)


def test_angle_from_rotmat(rotation_angles: torch.Tensor, rotation_matrices: torch.Tensor) -> None:
    """Test computation of rotation angles from rotation matrices."""
    angles, angles_sin, angles_cos = angle_from_rotmat(rotation_matrices)
    assert torch.allclose(angles, rotation_angles)
    assert torch.allclose(angles_sin, torch.sin(rotation_angles))
    assert torch.allclose(angles_cos, torch.cos(rotation_angles))


def test_rotvec_to_rotmat(rotation_vectors: torch.Tensor, rotation_matrices: torch.Tensor) -> None:
    """Test conversion of rotation vectors to rotation matrices."""
    rotmats = rotvec_to_rotmat(rotation_vectors)
    assert torch.allclose(rotmats, rotation_matrices)


def test_rotmat_to_rotvec(rotation_vectors: torch.Tensor, rotation_matrices: torch.Tensor) -> None:
    """Test conversion of rotation matrices to rotation vectors."""
    rotvecs = rotmat_to_rotvec(rotation_matrices)
    assert torch.allclose(rotvecs, rotation_vectors)


def test_rotquat_to_axis_angle(
    rotation_quaternions: torch.Tensor, rotation_vectors: torch.Tensor, tol: float = 1e-7
) -> None:
    """Test extraction of rotation angles and axes from unit quaternions."""
    rotation_angles = torch.norm(rotation_vectors, dim=-1)
    rotation_axes = torch.where(
        rotation_angles[:, None] < tol,
        torch.zeros_like(rotation_vectors),
        rotation_vectors / rotation_angles[:, None],
    )

    rotangs, rotaxs = _rotquat_to_axis_angle(rotation_quaternions)

    assert torch.allclose(rotangs, rotation_angles)
    assert torch.allclose(rotaxs, rotation_axes)


def test_rotquat_to_rotvec(
    rotation_quaternions: torch.Tensor, rotation_vectors: torch.Tensor
) -> None:
    """Test conversion of unit quaternions to rotation vectors."""
    rotvecs = rotquat_to_rotvec(rotation_quaternions)
    assert torch.allclose(rotvecs, rotation_vectors)


def test_rotquat_to_rotmat(
    rotation_quaternions: torch.Tensor, rotation_matrices: torch.Tensor
) -> None:
    """Test conversion of unit quaternions to rotation matrices."""
    rotmats = rotquat_to_rotmat(rotation_quaternions)
    assert torch.allclose(rotmats, rotation_matrices, atol=1e-5)


def test_apply_rotvec_to_rotmat(rotation_matrices: torch.Tensor, rotation_vectors: torch.Tensor):
    """Test composition of rotation matrix and vector."""
    compound_rots = apply_rotvec_to_rotmat(rotation_matrices, rotation_vectors)
    compound_rotation = torch.einsum(
        "...ij,...jk->...ik", rotation_matrices, rotvec_to_rotmat(rotation_vectors)
    )
    assert torch.allclose(compound_rots, compound_rotation)


def _assert_rotation_matrix(matrices: torch.Tensor, tol: float = 1e-6) -> None:
    """Check if a batch of matrices are valid rotation matrices."""
    # Number of dimensions.
    assert matrices.ndim >= 2

    # 3 x 3 shape for SO(3).
    matrix_dims = matrices.shape[-2:]
    assert matrix_dims == (3, 3)

    # Check inverse equals transpose (U.T U = Id).
    mtm = torch.einsum("b...ij,b...ik->b...jk", matrices, matrices)
    id3 = _broadcast_identity(mtm)
    assert torch.allclose(mtm, id3, atol=tol)

    # Check determinants equal 1.
    assert torch.allclose(
        torch.linalg.det(matrices),
        torch.tensor(1.0, device=matrices.device, dtype=matrices.dtype),
        atol=tol,
    )


def test_scale_rotmat(rotation_matrices, tol=1e-7):
    """Test for scaling rotation matrices by scalars."""
    scales = torch.rand(rotation_matrices.shape[0], 1).to(rotation_matrices)
    angles_0, _, _ = angle_from_rotmat(rotation_matrices)
    scaled_mat = scale_rotmat(rotation_matrices, scales)

    # Scaled matrices should be rotation matrices.
    _assert_rotation_matrix(scaled_mat)

    # Angles should be scaled accordingly.
    angles_1, _, _ = angle_from_rotmat(scaled_mat)
    assert torch.allclose(angles_1 / (angles_0 + tol), scales.squeeze(-1), atol=1e-4)


def test_igso3_derivative(rotation_angles, lower=2e-1, l_max=1000, tol: float = 1e-7):
    """Test implementation of IGSO(3) expansion derivative."""
    # Generate sigma values for testing.
    sigma = torch.clamp(torch.rand(rotation_angles.shape[0]), min=lower, max=0.9)

    # Generate grid for expansions.
    l_grid = torch.arange(l_max + 1)

    # Enable grad for derivatives.
    rotangs = rotation_angles.clone()
    rotangs.requires_grad = True

    # Compute grad using autograd.
    digso3_auto = torch.autograd.grad(
        igso3_expansion(rotangs, sigma, l_grid, tol=tol).sum(), rotangs
    )[0]

    # Compare to manual implementation (tolerance needs to be increased due to numerics).
    digso3 = digso3_expansion(rotangs, sigma, l_grid, tol=tol)

    assert torch.allclose(digso3, digso3_auto, atol=1e-3, rtol=1e-3)


def test_dlog_igso3_derivative(rotation_angles, lower=2e-1, l_max=1000, tol: float = 1e-7):
    """Test derivative of the logarithm of the IGSO(3) expansion."""
    # Generate sigma values for testing.
    sigma = torch.clamp(torch.rand(rotation_angles.shape[0]), min=lower, max=0.9)

    # Generate grid for expansions.
    l_grid = torch.arange(l_max + 1)

    # Enable grad for derivatives.
    rotangs = rotation_angles.clone()
    rotangs.requires_grad = True

    # Compute grad using autograd.
    dlog_igso3_auto = torch.autograd.grad(
        torch.log(torch.abs(igso3_expansion(rotangs, sigma, l_grid, tol=tol)) + tol).sum(), rotangs
    )[0]

    # Compare to manual implementation (tolerance needs to be increased due to numerics).
    dlog_igso3 = dlog_igso3_expansion(rotangs, sigma, l_grid, tol=tol)

    assert torch.allclose(dlog_igso3, dlog_igso3_auto, atol=1e-3, rtol=1e-3)


def test_sample_igso3(
    num_omega: int = NUM_OMEGA,
    num_sigma: int = NUM_SIGMA,
    num_samples: int = NUM_SAMPLES,
    lower: float = 1e-4,
) -> None:
    """
    Test IGSO(3) sampling routine. Only shapes and whether valid rotation matrices are returned are
    checked.
    """

    # Construct sigma grid and samples.
    sigma_grid = torch.linspace(lower, 1.0, num_sigma)
    sigma = torch.clamp(torch.rand(num_sigma), min=lower)

    sampler = SampleIGSO3(
        num_omega=num_omega,
        sigma_grid=sigma_grid,
        l_max=1000,
    )

    # Check if lookup table has correct dimensions
    assert sampler.sigma_grid.shape == (num_sigma,)
    assert sampler.omega_grid.shape == (num_omega,)
    assert sampler.cdf_igso3.shape == (num_sigma, num_omega)

    # Sample and check shape and whether proper rotation matrices were returned.
    samples = sampler.sample(sigma, num_samples)
    assert samples.shape == (num_sigma, num_samples, 3, 3)
    _assert_rotation_matrix(samples)

    # Check sampling of rotation axes for proper shape and unit length.
    samples_vectors = sampler.sample_vector(num_sigma, num_samples)
    samples_vectors_norm = torch.norm(samples_vectors, dim=-1)
    assert samples_vectors.shape == (num_sigma, num_samples, 3)
    assert torch.allclose(samples_vectors_norm, torch.ones_like(samples_vectors_norm))

    # Check shape of sampled angles.
    samples_angles = sampler.sample_angle(sigma, num_samples)
    assert samples_angles.shape == (num_sigma, num_samples)


def test_sample_uso3(
    num_omega: int = NUM_OMEGA,
    num_sigma: int = NUM_SIGMA,
    num_samples: int = NUM_SAMPLES,
    lower: float = 1e-4,
) -> None:
    """
    Test uniform SO(3) sampling routine. Only shapes and whether valid rotation matrices are
    returned are checked.
    """

    # Construct sigma grid and samples.
    sigma_grid = torch.linspace(lower, 1.0, num_sigma)
    sigma = torch.clamp(torch.rand(num_sigma), min=lower)

    sampler = SampleUSO3(
        num_omega=num_omega,
        sigma_grid=sigma_grid,
    )

    # Check if lookup table has correct dimensions
    assert sampler.sigma_grid.shape == (num_sigma,)
    assert sampler.omega_grid.shape == (num_omega,)
    assert sampler.cdf_igso3.shape == (1, num_omega)

    # Sample and check shape and whether proper rotation matrices were returned.
    samples = sampler.sample(sigma, num_samples)
    assert samples.shape == (num_sigma, num_samples, 3, 3)
    _assert_rotation_matrix(samples)

    # Check sampling of rotation axes for proper shape and unit length.
    samples_vectors = sampler.sample_vector(num_sigma, num_samples)
    samples_vectors_norm = torch.norm(samples_vectors, dim=-1)
    assert samples_vectors.shape == (num_sigma, num_samples, 3)
    assert torch.allclose(samples_vectors_norm, torch.ones_like(samples_vectors_norm))

    # Check shape of sampled angles.
    samples_angles = sampler.sample_angle(sigma, num_samples)
    assert samples_angles.shape == (num_sigma, num_samples)

    # Check shape sampling for USO(3).
    samples = sampler.sample_shape(num_sigma, num_samples)
    assert samples.shape == (num_sigma, num_samples, 3, 3)
    _assert_rotation_matrix(samples)


def test_score_so3(
    rotation_vectors,
    num_omega: int = NUM_OMEGA,
    num_sigma: int = NUM_SIGMA,
    num_samples: int = NUM_SAMPLES,
    lower: float = 1e-1,
):
    """
    Test SO(3) score computation. Does not check for overall correctness of score, only for
    tensor shapes and whether derivatives are properly propagated through the score function.
    """
    # Construct sigma grid and samples.
    sigma_grid = torch.linspace(1e-3, 1.5, num_sigma)
    sigma = torch.clamp(torch.rand(num_samples), min=lower)

    score_so3 = ScoreSO3(
        num_omega=num_omega,
        sigma_grid=sigma_grid,
        l_max=2000,
        tol=1e-6,
    )

    # Test computation of score norm (shapes and general sanity only).
    score_scaling = score_so3.get_score_scaling(sigma)
    assert score_scaling.shape == sigma.shape
    assert torch.all(score_scaling >= 0.0)

    # Enable grad for rotation vectors.
    rotvecs = rotation_vectors.clone()
    rotvecs.requires_grad = True

    # Check whether score has the correct shape.
    score = score_so3(sigma, rotvecs)
    assert score.shape == rotvecs.shape
    # Check whether grad has the correct shape.
    dscore = torch.autograd.grad(score.sum(), rotvecs)[0]
    assert dscore.shape == rotvecs.shape



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
import os
from unittest.mock import patch

import hydra
import torch
import yaml

from bioemu.sample import generate_batch
from bioemu.shortcuts import CosineVPSDE, DiGSO3SDE

from .test_embeds import TEST_SEQ, mock_run_colabfold


# Write a score model mock that inputs a batch and t, and outputs a dict of pos and node_orientations as keys
def mock_score_model(batch, t):
    device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")
    return {
        "pos": torch.rand(batch["pos"].shape).to(device),
        "node_orientations": torch.rand(batch["node_orientations"].shape[0], 3).to(device),
    }


def test_generate_batch(tmp_path):
    sequence = TEST_SEQ
    sdes = {"node_orientations": DiGSO3SDE(), "pos": CosineVPSDE()}
    batch_size = 2
    seed = 42
    with open(
        os.path.join(os.path.dirname(__file__), "../src/bioemu/config/denoiser/dpm.yaml")
    ) as f:
        denoiser_config = yaml.safe_load(f)
    denoiser = hydra.utils.instantiate(denoiser_config)

    # Mock the run_colabfold function
    with patch("bioemu.get_embeds.run_colabfold", side_effect=mock_run_colabfold), patch(
        "bioemu.get_embeds.ensure_colabfold_install"
    ):
        # cache_embeds_dir could be None when input to get_colabfold_embeds
        batch = generate_batch(
            score_model=mock_score_model,
            sequence=sequence,
            sdes=sdes,
            batch_size=batch_size,
            seed=seed,
            denoiser=denoiser,
            cache_embeds_dir=None,
        )

    assert "pos" in batch
    assert "node_orientations" in batch
    assert batch["pos"].shape == (batch_size, len(sequence), 3)
    assert batch["node_orientations"].shape == (batch_size, len(sequence), 3, 3)



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
"""Script for sampling from a trained model."""

import logging
import os
import typing
from collections.abc import Callable
from pathlib import Path
from typing import Literal

import hydra
import numpy as np
import stackprinter
from huggingface_hub import hf_hub_download

stackprinter.set_excepthook(style="darkbg2")

import torch
import yaml
from torch_geometric.data.batch import Batch

from .chemgraph import ChemGraph
from .convert_chemgraph import save_pdb_and_xtc
from .get_embeds import get_colabfold_embeds
from .models import DiGConditionalScoreModel
from .sde_lib import SDE
from .seq_io import parse_sequence, write_fasta
from .utils import count_samples_in_output_dir, format_npz_samples_filename

logger = logging.getLogger(__name__)

DEFAULT_DENOISER_CONFIG_DIR = Path(__file__).parent / "config/denoiser/"
SupportedDenoisersLiteral = Literal["dpm", "heun"]
SUPPORTED_DENOISERS = list(typing.get_args(SupportedDenoisersLiteral))


def maybe_download_checkpoint(
    *,
    model_name: str | None,
    ckpt_path: str | Path | None = None,
    model_config_path: str | Path | None = None,
) -> tuple[str, str]:
    """If ckpt_path and model config_path are specified, return them, else download named model from huggingface.
    Returns:
        tuple[str, str]: path to checkpoint, path to model config
    """
    if ckpt_path is not None:
        assert model_config_path is not None, "Must provide model_config_path if ckpt_path is set."
        return str(ckpt_path), str(model_config_path)
    assert model_name is not None
    assert (
        model_config_path is None
    ), f"Named model {model_name} comes with its own config. Do not provide model_config_path."
    ckpt_path = hf_hub_download(
        repo_id="microsoft/bioemu", filename=f"checkpoints/{model_name}/checkpoint.ckpt"
    )
    model_config_path = hf_hub_download(
        repo_id="microsoft/bioemu", filename=f"checkpoints/{model_name}/config.yaml"
    )
    return str(ckpt_path), str(model_config_path)


@torch.no_grad()
def main(
    sequence: str | Path,
    num_samples: int,
    output_dir: str | Path,
    batch_size_100: int = 10,
    model_name: str | None = "bioemu-v1.0",
    ckpt_path: str | Path | None = None,
    model_config_path: str | Path | None = None,
    denoiser_type: SupportedDenoisersLiteral | None = "dpm",
    denoiser_config_path: str | Path | None = None,
    cache_embeds_dir: str | Path | None = None,
    msa_host_url: str | None = None,
    filter_samples: bool = True,
) -> None:
    """
    Generate samples for a specified sequence, using a trained model.

    Args:
        sequence: Amino acid sequence for which to generate samples, or a path to a .fasta file, or a path to an .a3m file with MSAs.
            If it is not an a3m file, then colabfold will be used to generate an MSA and embedding.
        num_samples: Number of samples to generate. If `output_dir` already contains samples, this function will only generate additional samples necessary to reach the specified `num_samples`.
        output_dir: Directory to save the samples. Each batch of samples will initially be dumped as .npz files. Once all batches are sampled, they will be converted to .xtc and .pdb.
        batch_size_100: Batch size you'd use for a sequence of length 100. The batch size will be calculated from this, assuming
           that the memory requirement to compute each sample scales quadratically with the sequence length.
        model_name: Name of pretrained model to use. The model will be retrieved from huggingface. If not set,
           this defaults to `bioemu-v1.0`. If this is set, you do not need to provide `ckpt_path` or `model_config_path`.
        ckpt_path: Path to the model checkpoint. If this is set, `model_name` will be ignored.
        model_config_path: Path to the model config, defining score model architecture and the corruption process the model was trained with.
           Only required if `ckpt_path` is set.
        denoiser_type: Denoiser to use for sampling, if `denoiser_config_path` not specified. Comes in with default parameter configuration. Must be one of ['dpm', 'heun']
        denoiser_config_path: Path to the denoiser config, defining the denoising process.
        cache_embeds_dir: Directory to store MSA embeddings. If not set, this defaults to `COLABFOLD_DIR/embeds_cache`.
        msa_host_url: MSA server URL. If not set, this defaults to colabfold's remote server. If sequence is an a3m file, this is ignored.
        filter_samples: Filter out unphysical samples with e.g. long bond distances or steric clashes.
    """
    output_dir = Path(output_dir).expanduser().resolve()
    output_dir.mkdir(parents=True, exist_ok=True)  # Fail fast if output_dir is non-writeable

    ckpt_path, model_config_path = maybe_download_checkpoint(
        model_name=model_name, ckpt_path=ckpt_path, model_config_path=model_config_path
    )

    assert os.path.isfile(ckpt_path), f"Checkpoint {ckpt_path} not found"
    assert os.path.isfile(model_config_path), f"Model config {model_config_path} not found"

    with open(model_config_path) as f:
        model_config = yaml.safe_load(f)

    # User may have provided an MSA file instead of a sequence. This will be used for embeddings.
    msa_file = sequence if str(sequence).endswith(".a3m") else None

    if msa_file is not None and msa_host_url is not None:
        logger.warning(f"msa_host_url is ignored because MSA file {msa_file} is provided.")

    # Parse FASTA or A3M file if sequence is a file path. Extract the actual sequence.
    sequence = parse_sequence(sequence)

    fasta_path = output_dir / "sequence.fasta"
    if fasta_path.is_file():
        if parse_sequence(fasta_path) != sequence:
            raise ValueError(
                f"{fasta_path} already exists, but contains a sequence different from {sequence}!"
            )
    else:
        # Save FASTA file in output_dir
        write_fasta([sequence], fasta_path)

    model_state = torch.load(ckpt_path, map_location="cpu", weights_only=True)
    score_model: DiGConditionalScoreModel = hydra.utils.instantiate(model_config["score_model"])
    score_model.load_state_dict(model_state)
    sdes: dict[str, SDE] = hydra.utils.instantiate(model_config["sdes"])

    if denoiser_config_path is None:
        assert (
            denoiser_type in SUPPORTED_DENOISERS
        ), f"denoiser_type must be one of {SUPPORTED_DENOISERS}"
        denoiser_config_path = DEFAULT_DENOISER_CONFIG_DIR / f"{denoiser_type}.yaml"

    with open(denoiser_config_path) as f:
        denoiser_config = yaml.safe_load(f)
    denoiser = hydra.utils.instantiate(denoiser_config)

    logger.info(
        f"Sampling {num_samples} structures for sequence of length {len(sequence)} residues..."
    )
    batch_size = int(batch_size_100 * (100 / len(sequence)) ** 2)
    if batch_size == 0:
        logger.warning(f"Sequence {sequence} may be too long. Attempting with batch_size = 1.")
        batch_size = 1
    logger.info(f"Using batch size {min(batch_size, num_samples)}")

    existing_num_samples = count_samples_in_output_dir(output_dir)
    logger.info(f"Found {existing_num_samples} previous samples in {output_dir}.")
    for seed in range(existing_num_samples, num_samples, batch_size):
        n = min(batch_size, num_samples - seed)
        npz_path = output_dir / format_npz_samples_filename(seed, n)
        if npz_path.exists():
            raise ValueError(
                f"Not sure why {npz_path} already exists when so far only {existing_num_samples} samples have been generated."
            )
        logger.info(f"Sampling {seed=}")
        batch = generate_batch(
            score_model=score_model,
            sequence=sequence,
            sdes=sdes,
            batch_size=min(batch_size, n),
            seed=seed,
            denoiser=denoiser,
            cache_embeds_dir=cache_embeds_dir,
            msa_file=msa_file,
            msa_host_url=msa_host_url,
        )
        batch = {k: v.cpu().numpy() for k, v in batch.items()}
        np.savez(npz_path, **batch, sequence=sequence)

    logger.info("Converting samples to .pdb and .xtc...")
    samples_files = sorted(list(output_dir.glob("batch_*.npz")))
    sequences = [np.load(f)["sequence"].item() for f in samples_files]
    if set(sequences) != {sequence}:
        raise ValueError(f"Expected all sequences to be {sequence}, but got {set(sequences)}")
    positions = torch.tensor(np.concatenate([np.load(f)["pos"] for f in samples_files]))
    node_orientations = torch.tensor(
        np.concatenate([np.load(f)["node_orientations"] for f in samples_files])
    )
    save_pdb_and_xtc(
        pos_nm=positions,
        node_orientations=node_orientations,
        topology_path=output_dir / "topology.pdb",
        xtc_path=output_dir / "samples.xtc",
        sequence=sequence,
        filter_samples=filter_samples,
    )
    logger.info(f"Completed. Your samples are in {output_dir}.")


def generate_batch(
    score_model: torch.nn.Module,
    sequence: str,
    sdes: dict[str, SDE],
    batch_size: int,
    seed: int,
    denoiser: Callable,
    cache_embeds_dir: str | Path | None,
    msa_file: str | Path | None = None,
    msa_host_url: str | None = None,
) -> dict[str, torch.Tensor]:
    """Generate one batch of samples, using GPU if available.

    Args:
        score_model: Score model.
        sequence: Amino acid sequence.
        sdes: SDEs defining corruption process. Keys should be 'node_orientations' and 'pos'.
        embeddings_file: Path to embeddings file.
        batch_size: Batch size.
        seed: Random seed.
        msa_file: Optional path to an MSA A3M file.
        msa_host_url: MSA server URL for colabfold.
    """

    torch.manual_seed(seed)
    n = len(sequence)

    single_embeds_file, pair_embeds_file = get_colabfold_embeds(
        seq=sequence,
        cache_embeds_dir=cache_embeds_dir,
        msa_file=msa_file,
        msa_host_url=msa_host_url,
    )
    single_embeds = np.load(single_embeds_file)
    pair_embeds = np.load(pair_embeds_file)
    assert pair_embeds.shape[0] == pair_embeds.shape[1] == n
    assert single_embeds.shape[0] == n
    assert len(single_embeds.shape) == 2
    _, _, n_pair_feats = pair_embeds.shape  # [seq_len, seq_len, n_pair_feats]

    single_embeds, pair_embeds = torch.from_numpy(single_embeds), torch.from_numpy(pair_embeds)
    pair_embeds = pair_embeds.view(n**2, n_pair_feats)

    edge_index = torch.cat(
        [
            torch.arange(n).repeat_interleave(n).view(1, n**2),
            torch.arange(n).repeat(n).view(1, n**2),
        ],
        dim=0,
    )
    pos = torch.full((n, 3), float("nan"))
    node_orientations = torch.full((n, 3, 3), float("nan"))

    chemgraph = ChemGraph(
        edge_index=edge_index,
        pos=pos,
        node_orientations=node_orientations,
        single_embeds=single_embeds,
        pair_embeds=pair_embeds,
    )
    context_batch = Batch.from_data_list([chemgraph for _ in range(batch_size)])

    device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")
    sampled_chemgraph_batch = denoiser(
        sdes=sdes,
        device=device,
        batch=context_batch,
        score_model=score_model,
    )
    assert isinstance(sampled_chemgraph_batch, Batch)
    sampled_chemgraphs = sampled_chemgraph_batch.to_data_list()
    pos = torch.stack([x.pos for x in sampled_chemgraphs]).to("cpu")
    node_orientations = torch.stack([x.node_orientations for x in sampled_chemgraphs]).to("cpu")

    return {"pos": pos, "node_orientations": node_orientations}


if __name__ == "__main__":
    import logging

    import fire

    logging.basicConfig(level=logging.DEBUG)

    fire.Fire(main)



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

import os
from pathlib import Path

from Bio import SeqIO
from Bio.Seq import Seq
from Bio.SeqRecord import SeqRecord

StrPath = str | os.PathLike


def _ensure_seq_records(seqs: list[str | SeqRecord]) -> list[SeqRecord]:
    records = []
    for i, seq in enumerate(seqs):
        if isinstance(seq, str):
            records.append(SeqRecord(seq=Seq(seq), id=str(i)))
        else:
            records.append(seq)
    return records


def write_fasta(sequences: list[str | SeqRecord], fasta_file: StrPath) -> None:
    """Writes sequences in `seqs` in FASTA format

    Args:
        sequences (list[str | SeqRecord]): Sequences in 1-letter amino-acid code
        fasta_file (StrPath): Destination FASTA file
    """
    Path(fasta_file).parent.mkdir(parents=True, exist_ok=True)
    seq_records = _ensure_seq_records(sequences)
    with open(fasta_file, "w") as fasta_handle:
        SeqIO.write(seq_records, fasta_handle, format="fasta")


def read_fasta(fasta_file: StrPath) -> list[SeqRecord]:
    with open(fasta_file) as f:
        return list(SeqIO.parse(f, "fasta"))


def parse_sequence(sequence: StrPath) -> str:
    """Parse sequence if sequence is a file path. Otherwise just return the input."""
    try:
        if Path(sequence).is_file():
            # The same parser applies to both fasta and a3m files.
            rec = read_fasta(sequence)[0]
            return str(rec.seq)
    except OSError:
        # is_file() failed because the file name is too long.
        pass
    return str(sequence)



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
import abc
import logging
import os
import warnings
from collections.abc import Callable

import numpy as np
import torch
import torch.nn as nn
from torch._prims_common import DeviceLikeType
from tqdm import tqdm

from .sde_lib import SDE, _broadcast_like, maybe_expand

logger = logging.getLogger(__name__)


class SO3SDE(SDE, torch.nn.Module):
    def __init__(
        self,
        eps_t: float = 1e-4,
        num_sigma: int = 1000,
        num_omega: int = 1000,
        omega_exponent: int = 3,
        l_max: int = 1000,
        tol: float = 1e-6,
        cache_dir: str | None = None,
        overwrite_cache: bool = False,
    ) -> None:
        """
        Basic SDE class for SO(3) sampling. Implements the following SDE for rotations $R$:

        .. math ::
            \\mathrm{d} \\mathbf{R}^{(t)} = g(t) \\mathrm{d}\\mathbf{B}_\\mathrm{SO(3)}^{(t)}

        where $\\mathbf{B}_\\mathrm{SO(3)}$ is the Brownian motion on SO(3) and $g(t) is the diffusion
        coefficient.

        Rotations are sampled using a pre-computed lookup table of the IGSO(3) distribution, which
        is based on the series expansion

        .. math ::
            f(\\omega) = \\sum_{l=0}^\\infty (2l+1) \\exp \\left(-l(l+1)\frac{\\sigma^2}{2}\right)\frac{\\sin((l+\tfrac{1}{2})\\omega)}{\\sin(\tfrac{\\omega}{2})}

        truncated after `l_max` terms and discretized on a grid of `num_sigma` noise levels and
        `num_omega` angle values.

        Due to the manipulated objects being rotations in SO(3), input tensors x will always be of
        the general shape [... x 3] which should stay preserved under manipulations.

        Args:
            eps_t (float): smallest timestep used during the diffusion process.
            num_sigma (int, optional): Number of samples used to discretize the std dev dimension
              of the IGSO(3) lookup table. Defaults to 1000.
            num_omega (int, optional): Number of samples used to discretize the angle dimension of
              the IGSO(3) lookup table. Defaults to 1000.
            omega_exponent (int, optional): Make the angle grid denser for smaller angles by taking
              its power with the provided number. Defaults to 3.
            l_max (int, optional): Maximum number of terms used in the series expansion.  Defaults
              to 5000.
            tol: small offset for numerical stability.
            cache_dir: Path to an optional cache directory. If set to None, lookup tables are
              computed on the fly.
            overwrite_cache: If set to true, existing cache files are overwritten. Can be used for
              updating stale caches.

        References
        ----------
        .. [#yim2023_1] Yim, Trippe, De Bortoli, Mathieu, Doucet, Barzilay, Jaakkola:
            SE(3) diffusion model with application to protein backbone generation.
            arXiv preprint arXiv:2302.02277. 2023.
        """
        super().__init__()
        # Initialize grid of standard deviations (sigma) for sampling SO(3) distributions.
        sigma_grid = self._marginal_std(torch.linspace(eps_t, self.T, num_sigma))
        self.tol = tol

        if "AMLT_EXPERIMENT_NAME" in os.environ and (
            "RANK" in os.environ or "OMPI_COMM_WORLD_RANK" in os.environ
        ):
            warnings.warn("Set cache_dir of SO3SDE to None for AMLT runs.")
            cache_dir = None

        cache_dir = os.path.expanduser(cache_dir) if cache_dir is not None else None

        # Set up IGSO(3) sampling.
        self.igso3 = SampleIGSO3(
            num_omega=num_omega,
            sigma_grid=sigma_grid,
            omega_exponent=omega_exponent,
            l_max=l_max,
            tol=tol,
            cache_dir=cache_dir,
            overwrite_cache=overwrite_cache,
        )
        # Set up uniform SO(3) sampling (for prior).
        self.uso3 = SampleUSO3(
            num_omega=num_omega,
            sigma_grid=sigma_grid,
            omega_exponent=omega_exponent,
            tol=tol,
            cache_dir=cache_dir,
            overwrite_cache=overwrite_cache,
        )
        # Set up score function.
        self.score_function = ScoreSO3(
            num_omega=num_omega,
            sigma_grid=sigma_grid,
            omega_exponent=omega_exponent,
            l_max=l_max,
            tol=tol,
            cache_dir=cache_dir,
            overwrite_cache=overwrite_cache,
        )

    def compute_score(
        self,
        rotation_vectors: torch.Tensor,
        t: torch.Tensor,
        batch_idx: torch.LongTensor | None = None,
    ) -> torch.Tensor:
        """
        Compute current SO(3) score in vector representation (so(3)) given the rotation vectors
        corresponding to the perturbation and the current time scale.

        Args:
            rotation_vectors: Perturbation in vector form.
            t (torch.Tensor): Tensor of integration times.
            batch_idx (torch.LongTensor | None, optional): Index of batch assignments for sparse representations.
              Defaults to None

        Returns:
            SO(3) score in vector form.
        """
        sigma = self._marginal_std(t)
        sigma = maybe_expand(sigma, batch_idx)
        score = self.score_function(sigma, rotation_vectors)
        return score

    def get_score_scaling(
        self,
        t: torch.Tensor,
        batch_idx: torch.LongTensor | None = None,
    ) -> torch.Tensor:
        """
        Get the scaling of the SO(3) score in the loss function given the current perturbation time.

        Args:
            t (torch.Tensor): Tensor of integration times.
            batch_idx (B, optional): Index of batch assignments for sparse representations.
              Defaults to None

        Returns:
            Score scaling at time t.
        """
        sigma = self._marginal_std(t)
        score_scaling = self.score_function.get_score_scaling(sigma)
        score_scaling = maybe_expand(score_scaling, batch_idx)
        return score_scaling

    @abc.abstractmethod
    def beta(self, t: torch.Tensor) -> torch.Tensor:
        """This should be implemented to compute diffusion coefficient $g(t)$ in the SDE."""
        ...

    @abc.abstractmethod
    def _marginal_std(self, t: torch.Tensor) -> torch.Tensor:
        """This should be implemented to compute marginal standard deviation."""
        ...

    def sde(
        self, x: torch.Tensor, t: torch.Tensor, batch_idx: torch.LongTensor | None = None
    ) -> tuple[torch.Tensor, torch.Tensor]:
        """
        Differential equation corresponding to the process. The SO(3) diffusion has no drift. Care
        should be taken during integration to properly handle SO(3) peculiarities. E.g. scaling
        needs to be done in so(3) vector representation, action is matrix multiplication instead of
        addition. All returned properties are in vector representation.

        Args:
            x: Input tensor of rotation matrices.
            t: Tensor of integration times.
            batch_idx: Index of batch assignments for sparse representations.

        Returns:
            tuple[torch.Tensor, torch.Tensor]: Drift and diffusion associated with SDE.
        """
        beta_t = self.beta(t)
        # There is no drift in the SO(3) SDE (corresponds to zero vector).
        drift = torch.zeros_like(x[..., 0])
        diffusion = maybe_expand(beta_t, batch_idx, drift)
        return drift, diffusion

    def mean_coeff_and_std(
        self, x: torch.Tensor, t: torch.Tensor, batch_idx: torch.LongTensor | None = None
    ) -> tuple[torch.Tensor, torch.Tensor]:
        """Returns mean coefficient and standard deviation of marginal distribution at time t."""
        mean, std = self.marginal_prob(
            torch.ones_like(x[..., 0]), t, batch_idx
        )  # mean_coeff: vector form, std: same shape as x
        std = _broadcast_like(std, mean)
        return mean, std

    def prior_sampling(
        self,
        shape: torch.Size | tuple,
        device: DeviceLikeType | None = None,
    ) -> torch.Tensor:
        """
        Sample from the prior distribution, in this case the uniform distribution over SO(3).

        Args:
            shape: Shape of sample.

        Returns:
            torch.Tensor: Samples drawn from the SO(3) distribution.
        """
        # Check if input shape corresponds to rotation matrix
        if len(shape) > 2:
            assert shape[-2:] == (3, 3)
            # Remove rotation matrix dimensions (will get added again by the sampling).
            # B x ... x 3 x 3 -> B x ...
            shape = shape[:-2]

        # Check shape dimensions.
        assert len(shape) <= 2

        # If only one dimension is given, pad with auxiliary dimension. Final shape will be [B x F]
        # where F=1 if padded.
        if len(shape) == 1:
            remove_aux_dim = True
            shape = shape + (1,)
        else:
            remove_aux_dim = False

        # Sample from distribution in rotation matrix representation with the shape [B x F x 3 x 3].
        samples = self.uso3.sample_shape(*shape).to(device)

        # Remove auxiliary dimension F=1 if padded. The dim=-3 is due to the last two dimensions
        # always being the [3 x 3] rotation matrix components.
        if remove_aux_dim:
            # [B x 1 x 3 x 3] -> [B x 3 x 3]
            samples = samples.squeeze(-3)

        return samples

    @torch.no_grad()
    def sample_marginal(
        self, x: torch.Tensor, t: torch.Tensor, batch_idx: torch.LongTensor | None = None
    ) -> torch.Tensor:
        """
        Sample marginal for x(t) given x(0). For the IGSO(3) distribution, a sample with mean $\\mu$
        and variance $\nu$ can be obtained by decomposing the sampling process into

        .. math ::
            \\mathrm{IG}_\\mathrm{SO(3)}(\\mu, \nu) =  \\mathrm{IG}_\\mathrm{SO(3)}(\\mu, 0) \times \\mathrm{IG}_\\mathrm{SO(3)}(1, \nu)

        The first term is simply the rotation matrix $\\mu$ and the second term is a rotation sampled
        from IGSO(3) with the angle sampled according to $\nu$ and an uniformly distributed axis.


        Args:
            x (torch.Tensor): Input tensor.
            t (torch.Tensor): Tensor of integration times.
            batch_idx (B, optional): Index of batch assignments for sparse representations.
              Defaults to None

        Returns:
            torch.Tensor: Sample drawn from SO(3) in rotation matrix representation.
        """
        mean, std = self.marginal_prob(x=x, t=t, batch_idx=batch_idx)

        # std is already expanded over batch here and we need exactly 1 sample per batch
        # entry, since the whole frame should be rotated at the same time.
        # r -> [batch x 1 x 3 x 3]
        r = self.igso3.sample(std, num_samples=1)

        # Sampling from IGSO(3) with mean and std is computed as
        # IGSO3(mean, std) = IGSO3(mean, 0) @ IGSO3(1, std)
        # with IGSO3(mean, 0) = mean (and IGSO3(1, std) = r in our case)
        sample = torch.einsum("b...j,b...sjk->b...sk", mean, r)

        # Remove sample axis (1).
        sample = sample.squeeze(-2)

        return sample


class DiGSO3SDE(SO3SDE):
    def __init__(
        self,
        eps_t: float = 1e-4,
        num_sigma: int = 1000,
        num_omega: int = 2000,
        omega_exponent: int = 3,
        l_max: int = 2000,
        sigma_min: float = 0.02,
        sigma_max: float = 1.65,
        tol: float = 1e-7,
        cache_dir: str | None = None,
        overwrite_cache: bool = False,
    ):
        """
        SDE module for SO(3) diffusion based on the DiG paper:
        https://arxiv.org/abs/2306.05445
        Samples the IGSO(3) based on an interpolated lookup table. Uses a geometric noise schedule
        as described in the publication. Operates in rotation matrix (SO(3)) representation. The
        underlying SDE is variance exploding.

        NOTE: DiG has an issue where the IGSO3 density is missing a factor of 2. We do not correct
          for this here, and just use the values of sigma_min and sigma_max from the DiG code.

        Args:
            eps_t (float): smallest timestep used during the diffusion process.
            num_sigma (int, optional): Number of samples used to discretize the std dev dimension
              of the IGSO(3) lookup table
            num_omega (int, optional): Number of samples used to discretize the angle dimension of
              the IGSO(3) lookup table.
            omega_exponent (int, optional): Make the angle grid denser for smaller angles by taking
              its power with the provided number. Defaults to 3.
            l_max (int, optional): Maximum number of terms used in the series expansion.
            sigma_min (float, optional): Minimum of the noise schedule.
            sigma_max (float, optional): Maximum of the noise schedule.
            tol: small offset for numerical stability.
            cache_dir: Path to an optional cache directory. If set to None, lookup tables are
              computed on the fly.
            overwrite_cache: If set to true, existing cache files are overwritten. Can be used for
              updating stale caches.
        """
        self.sigma_min = sigma_min
        self.sigma_max = sigma_max

        super().__init__(
            eps_t=eps_t,
            num_sigma=num_sigma,
            num_omega=num_omega,
            omega_exponent=omega_exponent,
            l_max=l_max,
            tol=tol,
            cache_dir=cache_dir,
            overwrite_cache=overwrite_cache,
        )

    def beta(self, t: torch.Tensor) -> torch.Tensor:
        """
        Compute coefficient for the diffusion term of the SDE. Taken from Eqn 30 in Song et al:
        NOTE: DiG code has the 2 outside the square root instead of inside. Why? Here we use the
        expression in Song et al instead.

        .. math ::
            g(t) = \\sqrt{\frac{d}{dt} \\sigma^2(t)}.

        Args:
            t (torch.Tensor): Current sample times.

        Returns:
            torch.Tensor: Diffusion coefficient evaluated at the provided times.
        """
        sigma = self._marginal_std(t)
        beta = sigma * np.sqrt(2.0 * np.log(self.sigma_max / self.sigma_min))
        return beta

    def _marginal_std(self, t: torch.Tensor) -> torch.Tensor:
        """
        Standard deviation used in IGSO(3). Taken from Eqn 30 in Song et al.

        .. math ::
            g(t) = \\sqrt{\frac{d}{dt} \\sigma^2(t)}.

        Args:
            t (torch.Tensor): Current sample times.

        Returns:
            torch.Tensor: Sigma evaluated at the provided times.
        """
        sigma = self.sigma_min * (self.sigma_max / self.sigma_min) ** t
        return sigma

    def marginal_prob(
        self, x: torch.Tensor, t: torch.Tensor, batch_idx: torch.LongTensor | None = None
    ) -> tuple[torch.Tensor, torch.Tensor]:
        """
        Compute distribution parameters for the marginal SO(3) sampling process.

        Args:
            x (torch.Tensor): Input tensor.
            t (torch.Tensor): Tensor of integration times.
            batch_idx (B, optional): Index of batch assignments for sparse representations.
              Defaults to None

        Returns:
            tuple[torch.Tensor, torch.Tensor]: Mean and standard deviation of target distribution.
        """
        # In FrameDiff, mean_coeff only used to sample stdev, no scaling of mean performed.
        # The underlying differential equation is variance exploding.
        # assert x.shape[-2:] == (3, 3) and len(x.shape) == 3
        std = self._marginal_std(t)
        # raise Exception("this broadcast is borked")
        # Broadcast to batch but not to dimensions of x.
        std = maybe_expand(std, batch_idx)
        return x, std


def scale_rotmat(
    rotation_matrix: torch.Tensor, scalar: torch.Tensor, tol: float = 1e-7
) -> torch.Tensor:
    """
    Scale rotation matrix. This is done by converting it to vector representation,
    scaling the length of the vector and converting back to matrix representation.

    Args:
        rotation_matrix: Rotation matrices.
        scalar: Scalar values used for scaling. Should have one fewer dimension than the
            rotation matrices for correct broadcasting.
        tol: Numerical offset for stability.

    Returns:
        Scaled rotation matrix.
    """
    # Check whether dimensions match.
    assert rotation_matrix.ndim - 1 == scalar.ndim
    scaled_rmat = rotvec_to_rotmat(rotmat_to_rotvec(rotation_matrix) * scalar, tol=tol)
    return scaled_rmat


def _broadcast_identity(target: torch.Tensor) -> torch.Tensor:
    """
    Generate a 3 by 3 identity matrix and broadcast it to a batch of target matrices.

    Args:
        target (torch.Tensor): Batch of target 3 by 3 matrices.

    Returns:
        torch.Tensor: 3 by 3 identity matrices in the shapes of the target.
    """
    id3 = torch.eye(3, device=target.device, dtype=target.dtype)
    id3 = torch.broadcast_to(id3, target.shape)
    return id3


def skew_matrix_exponential_map_axis_angle(
    angles: torch.Tensor, skew_matrices: torch.Tensor
) -> torch.Tensor:
    """
    Compute the matrix exponential of a rotation in axis-angle representation with the axis in skew
    matrix representation form. Maps the rotation from the lie group to the rotation matrix
    representation. Uses Rodrigues' formula instead of `torch.linalg.matrix_exp` for better
    computational performance:

    .. math::

        \\exp(\theta \\mathbf{K}) = \\mathbf{I} + \\sin(\theta) \\mathbf{K} + [1 - \\cos(\theta)] \\mathbf{K}^2

    Args:
        angles (torch.Tensor): Batch of rotation angles.
        skew_matrices (torch.Tensor): Batch of rotation axes in skew matrix (lie so(3)) basis.

    Returns:
        torch.Tensor: Batch of corresponding rotation matrices.
    """
    # Set up identity matrix and broadcast.
    id3 = _broadcast_identity(skew_matrices)

    # Broadcast angle vector to right dimensions
    angles = angles[..., None, None]

    exp_skew = (
        id3
        + torch.sin(angles) * skew_matrices
        + (1.0 - torch.cos(angles))
        * torch.einsum("b...ik,b...kj->b...ij", skew_matrices, skew_matrices)
    )
    return exp_skew


def skew_matrix_exponential_map(
    angles: torch.Tensor, skew_matrices: torch.Tensor, tol=1e-7
) -> torch.Tensor:
    """
    Compute the matrix exponential of a rotation vector in skew matrix representation. Maps the
    rotation from the lie group to the rotation matrix representation. Uses the following form of
    Rodrigues' formula instead of `torch.linalg.matrix_exp` for better computational performance
    (in this case the skew matrix already contains the angle factor):

    .. math ::

        \\exp(\\mathbf{K}) = \\mathbf{I} + \frac{\\sin(\theta)}{\theta} \\mathbf{K} + \frac{1-\\cos(\theta)}{\theta^2} \\mathbf{K}^2

    This form has the advantage, that Taylor expansions can be used for small angles (instead of
    having to compute the unit length axis by dividing the rotation vector by small angles):

    .. math ::

        \frac{\\sin(\theta)}{\theta} \approx 1 - \frac{\theta^2}{6}
        \frac{1-\\cos(\theta)}{\theta^2} \approx \frac{1}{2} - \frac{\theta^2}{24}

    Args:
        angles (torch.Tensor): Batch of rotation angles.
        skew_matrices (torch.Tensor): Batch of rotation axes in skew matrix (lie so(3)) basis.

    Returns:
        torch.Tensor: Batch of corresponding rotation matrices.
    """
    # Set up identity matrix and broadcast.
    id3 = _broadcast_identity(skew_matrices)

    # Broadcast angles and pre-compute square.
    angles = angles[..., None, None]
    angles_sq = angles.square()

    # Get standard terms.
    sin_coeff = torch.sin(angles) / angles
    cos_coeff = (1.0 - torch.cos(angles)) / angles_sq
    # Use second order Taylor expansion for values close to zero.
    sin_coeff_small = 1.0 - angles_sq / 6.0
    cos_coeff_small = 0.5 - angles_sq / 24.0

    mask_zero = torch.abs(angles) < tol
    sin_coeff = torch.where(mask_zero, sin_coeff_small, sin_coeff)
    cos_coeff = torch.where(mask_zero, cos_coeff_small, cos_coeff)

    # Compute matrix exponential using Rodrigues' formula.
    exp_skew = (
        id3
        + sin_coeff * skew_matrices
        + cos_coeff * torch.einsum("b...ik,b...kj->b...ij", skew_matrices, skew_matrices)
    )
    return exp_skew


def rotvec_to_rotmat(rotation_vectors: torch.Tensor, tol: float = 1e-7) -> torch.Tensor:
    """
    Convert rotation vectors to rotation matrix representation. The length of the rotation vector
    is the angle of rotation, the unit vector the rotation axis.

    Args:
        rotation_vectors (torch.Tensor): Batch of rotation vectors.
        tol: small offset for numerical stability.

    Returns:
        torch.Tensor: Rotation in rotation matrix representation.
    """
    # Compute rotation angle as vector norm.
    rotation_angles = torch.norm(rotation_vectors, dim=-1)

    # Map axis to skew matrix basis.
    skew_matrices = vector_to_skew_matrix(rotation_vectors)

    # Compute rotation matrices via matrix exponential.
    rotation_matrices = skew_matrix_exponential_map(rotation_angles, skew_matrices, tol=tol)

    return rotation_matrices


def rotmat_to_rotvec(rotation_matrices: torch.Tensor) -> torch.Tensor:
    """
    Convert a batch of rotation matrices to rotation vectors (logarithmic map from SO(3) to so(3)).
    The standard logarithmic map can be derived from Rodrigues' formula via Taylor approximation
    (in this case operating on the vector coefficients of the skew so(3) basis).

    ..math ::

        \\left[\\log(\\mathbf{R})\right]^\\lor = \frac{\theta}{2\\sin(\theta)} \\left[\\mathbf{R} - \\mathbf{R}^\top\right]^\\lor

    This formula has problems at 1) angles theta close or equal to zero and 2) at angles close and
    equal to pi.

    To improve numerical stability for case 1), the angle term at small or zero angles is
    approximated by its truncated Taylor expansion:

    .. math ::

        \\left[\\log(\\mathbf{R})\right]^\\lor \approx \frac{1}{2} (1 + \frac{\theta^2}{6}) \\left[\\mathbf{R} - \\mathbf{R}^\top\right]^\\lor

    For angles close or equal to pi (case 2), the outer product relation can be used to obtain the
    squared rotation vector:

    .. math :: \\omega \\otimes \\omega = \frac{1}{2}(\\mathbf{I} + R)

    Taking the root of the diagonal elements recovers the normalized rotation vector up to the signs
    of the component. The latter can be obtained from the off-diagonal elements.

    Adapted from https://github.com/jasonkyuyim/se3_diffusion/blob/2cba9e09fdc58112126a0441493b42022c62bbea/data/so3_utils.py
    which was adapted from https://github.com/geomstats/geomstats/blob/master/geomstats/geometry/special_orthogonal.py
    with heavy help from https://cvg.cit.tum.de/_media/members/demmeln/nurlanov2021so3log.pdf

    Args:
        rotation_matrices (torch.Tensor): Input batch of rotation matrices.

    Returns:
        torch.Tensor: Batch of rotation vectors.
    """
    # Get angles and sin/cos from rotation matrix.
    angles, angles_sin, _ = angle_from_rotmat(rotation_matrices)
    # Compute skew matrix representation and extract so(3) vector components.
    vector = skew_matrix_to_vector(rotation_matrices - rotation_matrices.transpose(-2, -1))

    # Three main cases for angle theta, which are captured
    # 1) Angle is 0 or close to zero -> use Taylor series for small values / return 0 vector.
    mask_zero = torch.isclose(angles, torch.zeros_like(angles)).to(angles.dtype)
    # 2) Angle is close to pi -> use outer product relation.
    mask_pi = torch.isclose(angles, torch.full_like(angles, np.pi), atol=1e-2).to(angles.dtype)
    # 3) Angle is unproblematic -> use the standard formula.
    mask_else = (1 - mask_zero) * (1 - mask_pi)

    # Compute case dependent pre-factor (1/2 for angle close to 0, angle otherwise).
    numerator = mask_zero / 2.0 + angles * mask_else
    # The Taylor expansion used here is actually the inverse of the Taylor expansion of the inverted
    # fraction sin(x) / x which gives better accuracy over a wider range (hence the minus and
    # position in denominator).
    denominator = (
        (1.0 - angles**2 / 6.0) * mask_zero  # Taylor expansion for small angles.
        + 2.0 * angles_sin * mask_else  # Standard formula.
        + mask_pi  # Avoid zero division at angle == pi.
    )
    prefactor = numerator / denominator
    vector = vector * prefactor[..., None]

    # For angles close to pi, derive vectors from their outer product (ww' = 1 + R).
    id3 = _broadcast_identity(rotation_matrices)
    skew_outer = (id3 + rotation_matrices) / 2.0
    # Ensure diagonal is >= 0 for square root (uses identity for masking).
    # TODO: Investigate if this line can be removed entirely and replaced by the clamp operation
    # only.
    skew_outer = skew_outer + (torch.relu(skew_outer) - skew_outer) * id3

    # Get basic rotation vector as sqrt of diagonal (is unit vector).
    vector_pi = torch.sqrt(
        torch.clamp(torch.diagonal(skew_outer, dim1=-2, dim2=-1), min=1e-8)
    )  # Clamp to avoid nan gradient.

    # Compute the signs of vector elements (up to a global phase).
    # Fist select indices for outer product slices with the largest norm.
    signs_line_idx = torch.argmax(torch.norm(skew_outer, dim=-1), dim=-1).long()
    # Select rows of outer product and determine signs.
    signs_line = torch.take_along_dim(skew_outer, dim=-2, indices=signs_line_idx[..., None, None])
    signs_line = signs_line.squeeze(-2)
    signs = torch.sign(signs_line)

    # Apply signs and rotation vector.
    vector_pi = vector_pi * angles[..., None] * signs

    # Fill entries for angle == pi in rotation vector (basic vector has zero entries at this point).
    vector = vector + vector_pi * mask_pi[..., None]

    return vector


def angle_from_rotmat(
    rotation_matrices: torch.Tensor,
) -> tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
    """
    Compute rotation angles (as well as their sines and cosines) encoded by rotation matrices.
    Uses atan2 for better numerical stability for small angles.

    Args:
        rotation_matrices (torch.Tensor): Batch of rotation matrices.

    Returns:
        tuple[torch.Tensor, torch.Tensor, torch.Tensor]: Batch of computed angles, sines of the
          angles and cosines of angles.
    """
    # Compute sine of angles (uses the relation that the unnormalized skew vector generated by a
    # rotation matrix has the length 2*sin(theta))
    skew_matrices = rotation_matrices - rotation_matrices.transpose(-2, -1)
    skew_vectors = skew_matrix_to_vector(skew_matrices)
    angles_sin = torch.norm(skew_vectors, dim=-1) / 2.0
    # Compute the cosine of the angle using the relation cos theta = 1/2 * (Tr[R] - 1)
    angles_cos = (torch.einsum("...ii", rotation_matrices) - 1.0) / 2.0

    # Compute angles using the more stable atan2
    angles = torch.atan2(angles_sin, angles_cos)

    return angles, angles_sin, angles_cos


def vector_to_skew_matrix(vectors: torch.Tensor) -> torch.Tensor:
    """
    Map a vector into the corresponding skew matrix so(3) basis.
    ```
                [  0 -z  y]
    [x,y,z] ->  [  z  0 -x]
                [ -y  x  0]
    ```

    Args:
        vectors (torch.Tensor): Batch of vectors to be mapped to skew matrices.

    Returns:
        torch.Tensor: Vectors in skew matrix representation.
    """
    # Generate empty skew matrices.
    skew_matrices = torch.zeros((*vectors.shape, 3), device=vectors.device, dtype=vectors.dtype)

    # Populate positive values.
    skew_matrices[..., 2, 1] = vectors[..., 0]
    skew_matrices[..., 0, 2] = vectors[..., 1]
    skew_matrices[..., 1, 0] = vectors[..., 2]

    # Generate skew symmetry.
    skew_matrices = skew_matrices - skew_matrices.transpose(-2, -1)

    return skew_matrices


def skew_matrix_to_vector(skew_matrices: torch.Tensor) -> torch.Tensor:
    """
    Extract a rotation vector from the so(3) skew matrix basis.

    Args:
        skew_matrices (torch.Tensor): Skew matrices.

    Returns:
        torch.Tensor: Rotation vectors corresponding to skew matrices.
    """
    vectors = torch.zeros_like(skew_matrices[..., 0])
    vectors[..., 0] = skew_matrices[..., 2, 1]
    vectors[..., 1] = skew_matrices[..., 0, 2]
    vectors[..., 2] = skew_matrices[..., 1, 0]
    return vectors


def _rotquat_to_axis_angle(
    rotation_quaternions: torch.Tensor, tol: float = 1e-7
) -> tuple[torch.Tensor, torch.Tensor]:
    """
    Auxiliary routine for computing rotation angle and rotation axis from unit quaternions. To avoid
    complications, rotations vectors with angles below `tol` are set to zero.

    Args:
        rotation_quaternions (torch.Tensor): Rotation quaternions in [r, i, j, k] format.
        tol (float, optional): Threshold for small rotations. Defaults to 1e-7.

    Returns:
        tuple[torch.Tensor, torch.Tensor]: Rotation angles and axes.
    """
    # Compute rotation axis and normalize (accounting for small length axes).
    rotation_axes = rotation_quaternions[..., 1:]
    rotation_axes_norms = torch.norm(rotation_axes, dim=-1)

    # Compute rotation angle via atan2
    rotation_angles = 2.0 * torch.atan2(rotation_axes_norms, rotation_quaternions[..., 0])

    # Save division.
    rotation_axes = rotation_axes / (rotation_axes_norms[:, None] + tol)
    return rotation_angles, rotation_axes


def rotquat_to_rotvec(rotation_quaternions: torch.Tensor) -> torch.Tensor:
    """
    Convert unit quaternions to rotation vectors.

    Args:
        rotation_quaternions (torch.Tensor): Input quaternions in [r,i,j,k] format.

    Returns:
        torch.Tensor: Rotation vectors.
    """
    rotation_angles, rotation_axes = _rotquat_to_axis_angle(rotation_quaternions)
    rotation_vectors = rotation_axes * rotation_angles[..., None]
    return rotation_vectors


def rotquat_to_rotmat(rotation_quaternions: torch.Tensor) -> torch.Tensor:
    """
    Convert unit quaternion to rotation matrix.

    Args:
        rotation_quaternions (torch.Tensor): Input quaternions in [r,i,j,k] format.

    Returns:
        torch.Tensor: Rotation matrices.
    """
    rotation_angles, rotation_axes = _rotquat_to_axis_angle(rotation_quaternions)
    skew_matrices = vector_to_skew_matrix(rotation_axes * rotation_angles[..., None])
    rotation_matrices = skew_matrix_exponential_map(rotation_angles, skew_matrices)
    return rotation_matrices


def apply_rotvec_to_rotmat(
    rotation_matrices: torch.Tensor,
    rotation_vectors: torch.Tensor,
    tol: float = 1e-7,
) -> torch.Tensor:
    """
    Update a rotation encoded in a rotation matrix with a rotation vector.

    Args:
        rotation_matrices: Input batch of rotation matrices.
        rotation_vectors: Input batch of rotation vectors.
        tol: Small offset for numerical stability.

    Returns:
        Updated rotation matrices.
    """
    # Convert vector to matrices.
    rmat_right = rotvec_to_rotmat(rotation_vectors, tol=tol)
    # Accumulate rotation.
    rmat_rotated = torch.einsum("...ij,...jk->...ik", rotation_matrices, rmat_right)
    return rmat_rotated


def rotmat_to_skew_matrix(mat: torch.Tensor) -> torch.Tensor:
    """
    Generates skew matrix for corresponding rotation matrix.

    Args:
        mat (torch.Tensor): Batch of rotation matrices.

    Returns:
        torch.Tensor: Skew matrices in the shapes of mat.
    """
    vec = rotmat_to_rotvec(mat)
    return vector_to_skew_matrix(vec)


def skew_matrix_to_rotmat(skew: torch.Tensor) -> torch.Tensor:
    """
    Generates rotation matrix for corresponding skew matrix.

    Args:
        skew (torch.Tensor): Batch of target 3 by 3 skew symmetric matrices.

    Returns:
        torch.Tensor: Rotation matrices in the shapes of skew.
    """
    vec = skew_matrix_to_vector(skew)
    return rotvec_to_rotmat(vec)


def local_log(point: torch.Tensor, base_point: torch.Tensor) -> torch.Tensor:
    """
    Matrix logarithm. Computes left-invariant vector field of beinging base_point to point
    on the manifold. Follows the signature of geomstats' equivalent function.
    https://geomstats.github.io/api/geometry.html#geomstats.geometry.lie_group.MatrixLieGroup.log

    Args:
        point (torch.Tensor): Batch of rotation matrices to compute vector field at.
        base_point (torch.Tensor): Transport coordinates to take matrix logarithm.

    Returns:
        torch.Tensor: Skew matrix that holds the vector field (in the tangent space).
    """
    return rotmat_to_skew_matrix(rot_mult(rot_transpose(base_point), point))


def multidim_trace(mat: torch.Tensor) -> torch.Tensor:
    """Take the trace of a matrix with leading dimensions."""
    return torch.einsum("...ii->...", mat)


def geodesic_dist(mat_1: torch.Tensor, mat_2: torch.Tensor) -> torch.Tensor:
    """
    Calculate the geodesic distance of two rotation matrices.

    Args:
        mat_1 (torch.Tensor): First rotation matrix.
        mat_2 (torch.Tensor): Second rotation matrix.

    Returns:
        Scalar for the geodesic distance between mat_1 and mat_2 with the same
        leading (i.e. batch) dimensions.
    """
    A = rotmat_to_skew_matrix(rot_mult(rot_transpose(mat_1), mat_2))
    return torch.sqrt(multidim_trace(rot_mult(A, rot_transpose(A))))


def rot_transpose(mat: torch.Tensor) -> torch.Tensor:
    """Take the transpose of the last two dimensions."""
    return torch.transpose(mat, -1, -2)


def rot_mult(mat_1: torch.Tensor, mat_2: torch.Tensor) -> torch.Tensor:
    """Matrix multiply two rotation matrices with leading dimensions."""
    return torch.einsum("...ij,...jk->...ik", mat_1, mat_2)


def rot_vf(mat_t: torch.Tensor, mat_1: torch.Tensor) -> torch.Tensor:
    """
    Computes the vector field Log_{mat_t}(mat_1).

    Args:
        mat_t (torch.Tensor): base point to compute vector field at.
        mat_1 (torch.Tensor): target rotation.

    Returns:
        Rotation vector representing the vector field.
    """
    return rotmat_to_rotvec(rot_mult(rot_transpose(mat_t), mat_1))


def geodesic_t(t: float, mat: torch.Tensor, base_mat: torch.Tensor) -> torch.Tensor:
    """
    Computes the geodesic at time t. Specifically, R_t = Exp_{base_mat}(t * Log_{base_mat}(mat)).

    Args:
        t: time along geodesic.
        mat: target points on manifold.
        base_mat: source point on manifold.

    Returns:
        Point along geodesic starting at base_mat and ending at mat.
    """
    vec_vf = rot_vf(base_mat, mat)
    mat_t = rotvec_to_rotmat(t * vec_vf)

    if base_mat.shape != mat_t.shape:
        raise ValueError(f"Incompatible shapes: base_mat={base_mat.shape}, mat_t={mat_t.shape}")
    return torch.einsum("...ij,...jk->...ik", base_mat, mat_t)


class SO3LookupCache:
    def __init__(
        self,
        cache_dir: str,
        cache_file: str,
        overwrite: bool = False,
    ) -> None:
        """
        Auxiliary class for handling storage / loading of SO(3) lookup tables in npz format.

        Args:
            cache_dir: Path to the cache directory.
            cache_file: Basic file name of the cache file.
            overwrite: Whether existing cache files should be overwritten if requested.
        """
        if not cache_file.endswith(".npz"):
            raise ValueError("Filename should have '.npz' extension.")
        self.cache_file = cache_file
        self.cache_dir = cache_dir
        self.cache_path = os.path.join(cache_dir, cache_file)
        self.overwrite = overwrite

    @property
    def path_exists(self) -> bool:
        return os.path.exists(self.cache_path)

    @property
    def dir_exists(self) -> bool:
        return os.path.exists(self.cache_dir)

    def delete_cache(self) -> None:
        """
        Delete the cache file.
        """
        if self.path_exists:
            os.remove(self.cache_path)

    def load_cache(self) -> dict[str, torch.Tensor]:
        """
        Load data from the cache file.

        Returns:
            Dictionary of loaded data tensors.
        """
        if self.path_exists:
            # Load data and convert to torch tensors.
            npz_data = np.load(self.cache_path)
            torch_dict = {f: torch.from_numpy(npz_data[f]) for f in npz_data.files}
            logger.info(f"Data loaded from {self.cache_path}")
            return torch_dict
        else:
            raise ValueError(f"No cache data found at {self.cache_path}.")

    def save_cache(self, data: dict[str, torch.Tensor]) -> None:
        """
        Save a dictionary of tensors to the cache file. If overwrite is set to True, an existing
        file is overwritten, otherwise a warning is raised and the file is not modified.

        Args:
            data: Dictionary of tensors that should be saved to the cache.
        """
        if not self.dir_exists:
            os.makedirs(self.cache_dir)

        if self.path_exists:
            if self.overwrite:
                logger.info("Overwriting cache ...")
                self.delete_cache()
            else:
                logger.warn(
                    f"Cache at {self.cache_path} exits and overwriting disabled. Doing nothing."
                )
        else:
            # Move everything to CPU and numpy and store.
            logger.info(f"Data saved to {self.cache_path}")
            numpy_dict = {k: v.detach().cpu().numpy() for k, v in data.items()}
            np.savez(self.cache_path, **numpy_dict)


class BaseSampleSO3(nn.Module):
    so3_type: str = "base"  # cache basename

    def __init__(
        self,
        num_omega: int,
        sigma_grid: torch.Tensor,
        omega_exponent: int = 3,
        tol: float = 1e-7,
        interpolate: bool = True,
        cache_dir: str | None = None,
        overwrite_cache: bool = False,
    ) -> None:
        """
        Base torch.nn module for sampling rotations from the IGSO(3) distribution. Samples are
        created by uniformly sampling a rotation axis and using inverse transform sampling for
        the angles. The latter uses the associated SO(3) cumulative probability distribution
        function (CDF) and a uniform distribution [0,1] as described in [#leach2022_1]_. CDF values
        are obtained by numerically integrating the probability distribution evaluated on a grid of
        angles and noise levels and stored in a lookup table. Linear interpolation is used to
        approximate continuos sampling of the function. Angles are discretized in an interval [0,pi]
        and the grid can be squashed to have higher resolutions at low angles by taking different
        powers. Since sampling relies on tabulated values of the CDF and indexing in the form of
        `torch.bucketize`, gradients are not supported.

        Args:
            num_omega (int): Number of discrete angles used for generating the lookup table.
            sigma_grid (torch.Tensor): Grid of IGSO3 std devs.
            omega_exponent (int, optional): Make the angle grid denser for smaller angles by taking
              its power with the provided number. Defaults to 3.
            tol (float, optional): Small value for numerical stability. Defaults to 1e-7.
            interpolate (bool, optional): If enables, perform linear interpolation of the angle CDF
              to sample angles. Otherwise the closest tabulated point is returned. Defaults to True.
            cache_dir: Path to an optional cache directory. If set to None, lookup tables are
              computed on the fly.
            overwrite_cache: If set to true, existing cache files are overwritten. Can be used for
              updating stale caches.

        References
        ----------
        .. [#leach2022_1] Leach, Schmon, Degiacomi, Willcocks:
           Denoising diffusion probabilistic models on so (3) for rotational alignment.
           ICLR 2022 Workshop on Geometrical and Topological Representation Learning. 2022.
        """
        super().__init__()
        self.num_omega = num_omega
        self.omega_exponent = omega_exponent
        self.tol = tol
        self.interpolate = interpolate
        self.register_buffer("sigma_grid", sigma_grid, persistent=False)

        # Generate / load lookups and store in non-persistent buffers.
        omega_grid, cdf_igso3 = self._setup_lookup(sigma_grid, cache_dir, overwrite_cache)
        self.register_buffer("omega_grid", omega_grid, persistent=False)
        self.register_buffer("cdf_igso3", cdf_igso3, persistent=False)

    def _setup_lookup(
        self,
        sigma_grid: torch.Tensor,
        cache_dir: str | None = None,
        overwrite_cache: bool = False,
    ) -> tuple[torch.Tensor, torch.Tensor]:
        """
        Master function for setting up the lookup tables. These can either be loaded from a npz
        cache file or computed on the fly. Lookup tables will always be created and stored in double
        precision. Casting to the target dtype is done at the end of the function.

        Args:
            sigma_grid: Grid of sigma values used for computing the lookup tables.
            cache_dir: Path to the cache directory.
            overwrite_cache: If set to true, an existing cache is overwritten. Can be used for
              updating stale caches.

        Returns:
            Grid of angle values and SO(3) cumulative distribution function.
        """
        if cache_dir is not None:
            cache_name = self._get_cache_name()
            cache = SO3LookupCache(cache_dir, cache_name, overwrite=True)

            # If cache dir is provided, check whether the necessary cache exists and whether it
            # should be overwritten.
            if cache.path_exists and not overwrite_cache:
                # Load data from cache.
                cache_data = cache.load_cache()
                omega_grid = cache_data["omega_grid"]
                cdf_igso3 = cache_data["cdf_igso3"]
            else:
                # Store data in cache (overwrite if requested).
                omega_grid, cdf_igso3 = self._generate_lookup(sigma_grid)
                cache.save_cache({"omega_grid": omega_grid, "cdf_igso3": cdf_igso3})
        else:
            # Other wise just generate the tables.
            omega_grid, cdf_igso3 = self._generate_lookup(sigma_grid)

        return omega_grid.to(sigma_grid.dtype), cdf_igso3.to(sigma_grid.dtype)

    def _get_cache_name(self) -> str:
        """
        Auxiliary function for determining the cache file name based on the parameters (sigma,
        omega, l, etc.) used for generating the lookup tables.

        Returns:
            Base name of the cache file.
        """
        cache_name = f"cache_{self.so3_type:s}_s{torch.min(self.sigma_grid).cpu().item():04.3f}-{torch.max(self.sigma_grid).cpu().item():04.3f}-{self.sigma_grid.shape[0]:d}_o{self.num_omega:d}-{self.omega_exponent:d}.npz"
        return cache_name

    def get_sigma_idx(self, sigma: torch.Tensor) -> torch.Tensor:
        """
        Convert continuous sigmas to the indices of the closest tabulated values.

        Args:
            sigma (torch.Tensor): IGSO3 std devs.

        Returns:
            torch.Tensor: Index tensor mapping the provided sigma values to the internal lookup
              table.
        """
        return torch.bucketize(sigma, self.sigma_grid)

    def expansion_function(
        self, omega_grid: torch.Tensor, sigma_grid: torch.Tensor
    ) -> torch.Tensor:
        """
        Function for generating the angle probability distribution. Should return a 2D tensor with
        values for the std dev at the first dimension (rows) and angles at the second
        (columns).

        Args:
            omega_grid (torch.Tensor): Grid of angle values.
            sigma_grid (torch.Tensor): IGSO3 std devs.

        Returns:
            torch.Tensor: Distribution for angles discretized on a 2D grid.
        """
        raise NotImplementedError

    @torch.no_grad()
    def _generate_lookup(self, sigma_grid: torch.Tensor) -> tuple[torch.Tensor, torch.Tensor]:
        """
        Generate the lookup table for sampling from the target SO(3) CDF. The table is 2D, with the
        rows corresponding to different sigma values and the columns with angles computed on a grid.
        Variance is scaled by a factor of 1/2 to account for the deacceleration of time in the
        diffusion process due to the choice of SO(3) basis and guarantee time-reversibility (see
        appendix E.3 in [#yim2023_2]_). The returned tables are double precision and will be cast
        to the target dtype in `_setup_lookup`.

        Args:
            sigma_grid (torch.Tensor): Grid of IGSO3 std devs.

        Returns:
            tuple[torch.Tensor, torch.Tensor]: tuple containing the grid used to compute the angles
              and the associated lookup table.

        References
        ----------
        .. [#yim2023_2] Yim, Trippe, De Bortoli, Mathieu, Doucet, Barzilay, Jaakkola:
           SE(3) diffusion model with application to protein backbone generation.
           arXiv preprint arXiv:2302.02277. 2023.
        """

        current_device = sigma_grid.device
        sigma_grid_tmp = sigma_grid.to(torch.float64)

        # If cuda is available, initialize everything on GPU.
        # Even if Pytorch Lightning usually handles GPU allocation after initialization, this is
        # required to initialize the module in GPU reducing the initializaiton time by orders of magnitude.
        if torch.cuda.is_available():
            sigma_grid_tmp = sigma_grid_tmp.to(device="cuda")

        # Set up grid for angle resolution. Convert to double precision for better handling of numerics.
        omega_grid = torch.linspace(0.0, 1, self.num_omega + 1).to(sigma_grid_tmp)

        # If requested, increase sample density for lower values
        omega_grid = omega_grid**self.omega_exponent

        omega_grid = omega_grid * np.pi

        # Compute the expansion for all omegas and sigmas.
        pdf_igso3 = self.expansion_function(omega_grid, sigma_grid_tmp)

        # Apply the pre-factor from USO(3).
        pdf_igso3 = pdf_igso3 * (1.0 - torch.cos(omega_grid)) / np.pi

        # Compute the cumulative probability distribution.
        cdf_igso3 = integrate_trapezoid_cumulative(pdf_igso3, omega_grid)
        # Normalize integral area to 1.
        cdf_igso3 = cdf_igso3 / cdf_igso3[:, -1][:, None]

        # Move back to original device.
        cdf_igso3 = cdf_igso3.to(device=current_device)
        omega_grid = omega_grid.to(device=current_device)

        return omega_grid[1:].to(sigma_grid.dtype), cdf_igso3.to(sigma_grid.dtype)

    def sample(self, sigma: torch.Tensor, num_samples: int) -> torch.Tensor:
        """
        Generate samples from the target SO(3) distribution by sampling a rotation axis angle,
        which are then combined into a rotation vector and transformed into the corresponding
        rotation matrix via an exponential map.

        Args:
            sigma_indices (torch.Tensor): Indices of the IGSO3 std devs for which to take samples.
            num_samples (int): Number of angle samples to take for each std dev

        Returns:
            torch.Tensor: Sampled rotations in matrix representation with dimensions
              [num_sigma x num_samples x 3 x 3].
        """

        vectors = self.sample_vector(sigma.shape[0], num_samples)
        angles = self.sample_angle(sigma, num_samples)

        # Do postprocessing on angles.
        angles = self._process_angles(sigma, angles)

        rotation_vectors = vectors * angles[..., None]

        rotation_matrices = rotvec_to_rotmat(rotation_vectors, tol=self.tol)
        return rotation_matrices

    def _process_angles(self, sigma: torch.Tensor, angles: torch.Tensor) -> torch.Tensor:
        """
        Auxiliary function for performing additional processing steps on the sampled angles. One
        example would be to ensure sampled angles are 0 for a std dev of 0 for IGSO(3).

        Args:
            sigma (torch.Tensor): Current values of sigma.
            angles (torch.Tensor): Sampled angles.

        Returns:
            torch.Tensor: Processed sampled angles.
        """
        return angles

    def sample_vector(self, num_sigma: int, num_samples: int) -> torch.Tensor:
        """
        Uniformly sample rotation axis for constructing the overall rotation.

        Args:
            num_sigma (int): Number of samples to draw for each std dev.
            num_samples (int): Number of angle samples to take for each std dev.

        Returns:
            torch.Tensor: Batch of rotation axes with dimensions [num_sigma x num_samples x 3].
        """
        vectors = torch.randn(num_sigma, num_samples, 3, device=self.sigma_grid.device)
        vectors = vectors / torch.norm(vectors, dim=2, keepdim=True)
        return vectors

    def sample_angle(self, sigma: torch.Tensor, num_samples: int) -> torch.Tensor:
        """
        Create a series of samples from the IGSO(3) angle distribution.

        Args:
            sigma_indices (torch.Tensor): Indices of the IGSO3 std deves for which to
              take samples.
            num_samples (int): Number of angle samples to take for each std dev.

        Returns:
            torch.Tensor: Collected samples, will have the dimension [num_sigma x num_samples].
        """
        # Convert sigmas to respective indices for lookup table.
        sigma_indices = self.get_sigma_idx(sigma)
        # Get relevant sigma slices from stored CDFs.
        cdf_tmp = self.cdf_igso3[sigma_indices, :]

        # Draw from uniform distribution.
        p_uniform = torch.rand((*sigma_indices.shape, *[num_samples]), device=sigma_indices.device)

        # Determine indices for CDF.
        idx_stop = torch.sum(cdf_tmp[..., None] < p_uniform[:, None, :], dim=1).long()
        idx_start = torch.clamp(idx_stop - 1, min=0)

        if not self.interpolate:
            omega = torch.gather(cdf_tmp, dim=1, index=idx_stop)
        else:
            # Get CDF values.
            cdf_start = torch.gather(cdf_tmp, dim=1, index=idx_start)
            cdf_stop = torch.gather(cdf_tmp, dim=1, index=idx_stop)

            # Compute weights for linear interpolation.
            cdf_delta = torch.clamp(cdf_stop - cdf_start, min=self.tol)
            cdf_weight = torch.clamp((p_uniform - cdf_start) / cdf_delta, min=0.0, max=1.0)

            # Get angle range for interpolation.
            omega_start = self.omega_grid[idx_start]
            omega_stop = self.omega_grid[idx_stop]

            # Interpolate.
            omega = torch.lerp(omega_start, omega_stop, cdf_weight)

        return omega


class SampleIGSO3(BaseSampleSO3):
    so3_type = "igso3"  # cache basename

    def __init__(
        self,
        num_omega: int,
        sigma_grid: torch.Tensor,
        omega_exponent: int = 3,
        tol: float = 1e-7,
        interpolate: bool = True,
        l_max: int = 1000,
        cache_dir: str | None = None,
        overwrite_cache: bool = False,
    ) -> None:
        """
        Module for sampling rotations from the IGSO(3) distribution using the explicit series
        expansion.  Samples are created using inverse transform sampling based on the associated
        cumulative probability distribution function (CDF) and a uniform distribution [0,1] as
        described in [#leach2022_2]_. CDF values are obtained by numerically integrating the
        probability distribution evaluated on a grid of angles and noise levels and stored in a
        lookup table.  Linear interpolation is used to approximate continuos sampling of the
        function. Angles are discretized in an interval [0,pi] and the grid can be squashed to have
        higher resolutions at low angles by taking different powers.
        Since sampling relies on tabulated values of the CDF and indexing in the form of
        `torch.bucketize`, gradients are not supported.

        Args:
            num_omega (int): Number of discrete angles used for generating the lookup table.
            sigma_grid (torch.Tensor): Grid of IGSO3 std devs.
            omega_exponent (int, optional): Make the angle grid denser for smaller angles by taking
              its power with the provided number. Defaults to 3.
            tol (float, optional): Small value for numerical stability. Defaults to 1e-7.
            interpolate (bool, optional): If enables, perform linear interpolation of the angle CDF
              to sample angles. Otherwise the closest tabulated point is returned. Defaults to True.
            l_max (int, optional): Maximum number of terms used in the series expansion.
            cache_dir: Path to an optional cache directory. If set to None, lookup tables are
              computed on the fly.
            overwrite_cache: If set to true, existing cache files are overwritten. Can be used for
              updating stale caches.

        References
        ----------
        .. [#leach2022_2] Leach, Schmon, Degiacomi, Willcocks:
           Denoising diffusion probabilistic models on so (3) for rotational alignment.
           ICLR 2022 Workshop on Geometrical and Topological Representation Learning. 2022.
        """
        self.l_max = l_max
        super().__init__(
            num_omega=num_omega,
            sigma_grid=sigma_grid,
            omega_exponent=omega_exponent,
            tol=tol,
            interpolate=interpolate,
            cache_dir=cache_dir,
            overwrite_cache=overwrite_cache,
        )

    def _get_cache_name(self) -> str:
        """
        Auxiliary function for determining the cache file name based on the parameters (sigma,
        omega, l, etc.) used for generating the lookup tables.

        Returns:
            Base name of the cache file.
        """
        cache_name = f"cache_{self.so3_type:s}_s{torch.min(self.sigma_grid).cpu().item():04.3f}-{torch.max(self.sigma_grid).cpu().item():04.3f}-{self.sigma_grid.shape[0]:d}_l{self.l_max:d}_o{self.num_omega:d}-{self.omega_exponent:d}.npz"
        return cache_name

    def expansion_function(
        self,
        omega_grid: torch.Tensor,
        sigma_grid: torch.Tensor,
    ) -> torch.Tensor:
        """
        Use the truncated expansion of the IGSO(3) probability function to generate the lookup table.

        Args:
            omega_grid (torch.Tensor): Grid of angle values.
            sigma_grid (torch.Tensor): Grid of IGSO3 std devs.

        Returns:
            torch.Tensor: IGSO(3) distribution for angles discretized on a 2D grid.
        """
        return generate_igso3_lookup_table(omega_grid, sigma_grid, l_max=self.l_max, tol=self.tol)

    def _process_angles(self, sigma: torch.Tensor, angles: torch.Tensor) -> torch.Tensor:
        """
        Ensure sampled angles are 0 for small noise levels in IGSO(3). (Series expansion gives
        uniform probability distribution.)

        Args:
            sigma (torch.Tensor): Current values of sigma.
            angles (torch.Tensor): Sampled angles.

        Returns:
            torch.Tensor: Processed sampled angles.
        """
        angles = torch.where(
            sigma[..., None] < self.tol,
            torch.zeros_like(angles),
            angles,
        )
        return angles


class SampleUSO3(BaseSampleSO3):
    so3_type = "uso3"  # cache basename

    def __init__(
        self,
        num_omega: int,
        sigma_grid: torch.Tensor,
        omega_exponent: int = 3,
        tol: float = 1e-7,
        interpolate: bool = True,
        cache_dir: str | None = None,
        overwrite_cache: bool = False,
    ) -> None:
        """
        Module for sampling rotations from the USO(3) distribution. Can be used to generate initial
        unbiased samples in the reverse process.  Samples are created using inverse transform
        sampling based on the associated cumulative probability distribution function (CDF) and a
        uniform distribution [0,1] as described in [#leach2022_4]_. CDF values are obtained by
        numerically integrating the probability distribution evaluated on a grid of angles and noise
        levels and stored in a lookup table.  Linear interpolation is used to approximate continuos
        sampling of the function. Angles are discretized in an interval [0,pi] and the grid can be
        squashed to have higher resolutions at low angles by taking different powers.
        Since sampling relies on tabulated values of the CDF and indexing in the form of
        `torch.bucketize`, gradients are not supported.

        Args:
            num_omega (int): Number of discrete angles used for generating the lookup table.
            sigma_grid (torch.Tensor): Grid of IGSO3 std devs.
            omega_exponent (int, optional): Make the angle grid denser for smaller angles by taking
              its power with the provided number. Defaults to 3.
            tol (float, optional): Small value for numerical stability. Defaults to 1e-7.
            interpolate (bool, optional): If enables, perform linear interpolation of the angle CDF
              to sample angles. Otherwise the closest tabulated point is returned. Defaults to True.
            cache_dir: Path to an optional cache directory. If set to None, lookup tables are
              computed on the fly.
            overwrite_cache: If set to true, existing cache files are overwritten. Can be used for
              updating stale caches.

        References
        ----------
        .. [#leach2022_4] Leach, Schmon, Degiacomi, Willcocks:
           Denoising diffusion probabilistic models on so (3) for rotational alignment.
           ICLR 2022 Workshop on Geometrical and Topological Representation Learning. 2022.
        """
        super().__init__(
            num_omega=num_omega,
            sigma_grid=sigma_grid,
            omega_exponent=omega_exponent,
            tol=tol,
            interpolate=interpolate,
            cache_dir=cache_dir,
            overwrite_cache=overwrite_cache,
        )

    def get_sigma_idx(self, sigma: torch.Tensor) -> torch.Tensor:
        return torch.zeros_like(sigma).long()

    def sample_shape(self, num_sigma: int, num_samples: int) -> torch.Tensor:
        dummy_sigma = torch.zeros(num_sigma, device=self.sigma_grid.device)
        return self.sample(dummy_sigma, num_samples)

    def expansion_function(
        self,
        omega_grid: torch.Tensor,
        sigma_grid: torch.Tensor,
    ) -> torch.Tensor:
        """
        The probability density function of the uniform SO(3) distribution is the cosine scaling
        term (1-cos(omega))/pi which is applied automatically during sampling. This means, it is
        sufficient to return a tensor of ones to create the correct USO(3) lookup table.

        Args:
            omega_grid (torch.Tensor): Grid of angle values.
            sigma_grid (torch.Tensor): Grid of IGSO3 std devs.

        Returns:
            torch.Tensor: USO(3) distribution for angles discretized on a 2D grid.
        """
        return torch.ones(1, omega_grid.shape[0], device=omega_grid.device)


@torch.no_grad()
def integrate_trapezoid_cumulative(f_grid: torch.Tensor, x_grid: torch.Tensor) -> torch.Tensor:
    """
    Auxiliary function for numerically integrating a discretized 1D function using the trapezoid
    rule. This is mainly used for computing the cumulative probability distributions for sampling
    from the IGSO(3) distribution. Works on a single 1D grid or a batch of grids.

    Args:
        f_grid (torch.Tensor): Discretized function values.
        x_grid (torch.Tensor): Discretized input values.

    Returns:
        torch.Tensor: Integrated function (not normalized).
    """
    f_sum = f_grid[..., :-1] + f_grid[..., 1:]
    delta_x = torch.diff(x_grid, dim=-1)
    integral = torch.cumsum((f_sum * delta_x[None, :]) / 2.0, dim=-1)
    return integral


class ScoreSO3(nn.Module):
    def __init__(
        self,
        num_omega: int,
        sigma_grid: torch.Tensor,
        omega_exponent: int = 3,
        l_max: int = 1000,
        tol: float = 1e-7,
        cache_dir: str | None = None,
        overwrite_cache: bool = False,
    ) -> None:
        """
        Module for computing scores for SO(3) diffusion (see e.g. [#yim2023_1]_). Score computation
        is performed in the rotation vector representation (tangent space). The score is computed
        as:

        .. math ::
            \nabla_\\mathbf{q} = \frac{\\mathbf{q}}{|\\mathbf{q}|} \frac{\tfrac{\\partial}{\\partial |\\mathbf{q}|} f_{\\mathrm{IGSO(3)}}(|\\mathbf{q}|)}{f_{\\mathrm{IGSO(3)}}(|\\mathbf{q}|)}

        The fraction of the IGSO(3) derivative and probability function is obtained via truncated
        series expansion.

        In addition to the scores, the module also pre-computes score scaling values for the loss
        function according to the equation:

        .. math ::
            \frac{1}{\\lambda} = \\sqrt{\frac{\\sum_\\omega \\left[\\partial_\\omega \\log f_\\mathrm{IGSO(3)}(\\omega)\right]^2 f_\\mathrm{IGSO(3)}(\\omega)}{3 \\sum_\\omega f_\\mathrm{IGSO(3)}(\\omega)}}

        These values are stored in a lookup table and can be retrieved based on the closest
        tabulated noise levels. While the score supports full gradients during optimization, the
        scaling is static.

        In general, it is recommended to use the same grid settings as for sampling the rotations.

        Args:
            num_omega: Number of discrete angles used for generating the lookup table.
            sigma_grid: Grid of IGSO3 std devs.
            omega_exponent: Make the angle grid denser for smaller angles by taking its power with
              the provided number.
            l_max: Maximum number of terms used in the series expansion.
            tol: Small value for numerical stability.
            cache_dir: Path to an optional cache directory. If set to None, lookup tables are
              computed on the fly.
            overwrite_cache: If set to true, existing cache files are overwritten. Can be used for
              updating stale caches.

        References
        ----------
        .. [#yim2023_1] Yim, Trippe, De Bortoli, Mathieu, Doucet, Barzilay, Jaakkola:
           SE(3) diffusion model with application to protein backbone generation.
           arXiv preprint arXiv:2302.02277. 2023.
        """
        super().__init__()

        # Store different parameters and grids.
        self.l_max = l_max
        self.tol = tol
        self.num_omega = num_omega
        self.omega_exponent = omega_exponent
        self.register_buffer("l_grid", torch.arange(l_max + 1), persistent=False)
        self.register_buffer("sigma_grid", sigma_grid, persistent=False)

        # Pre-compute score scaling and store in buffer.
        score_scaling = self._setup_scaling(sigma_grid, cache_dir, overwrite_cache)
        self.register_buffer("score_scaling", score_scaling, persistent=False)

    def _setup_scaling(
        self,
        sigma_grid: torch.Tensor,
        cache_dir: str | None = None,
        overwrite_cache: bool = False,
    ) -> torch.Tensor:
        """
        Master function for setting up the lookup table for SO(3) score scaling. It can either be
        loaded from a npz cache file or computed on the fly.

        Args:
            sigma_grid: Grid of sigma values used for computing the lookup tables.
            cache_dir: Path to the cache directory.
            overwrite_cache: If set to true, an existing cache is overwritten. Can be used for
              updating stale caches.

        Returns:
            SO(3) score scaling.
        """
        if cache_dir is not None:
            cache_name = self._get_cache_name()
            cache = SO3LookupCache(cache_dir, cache_name, overwrite=True)

            # If cache dir is provided, check whether the necessary cache exists and whether it
            # should be overwritten.
            if cache.path_exists and not overwrite_cache:
                # Load data from cache.
                cache_data = cache.load_cache()
                score_scaling = cache_data["score_scaling"]
            else:
                # Store data in cache (overwrite if requested).
                score_scaling = self._compute_score_scaling(sigma_grid)
                cache.save_cache({"score_scaling": score_scaling})
        else:
            # Other wise just generate the tables.
            score_scaling = self._compute_score_scaling(sigma_grid)
        return score_scaling.to(sigma_grid.dtype)

    def _get_cache_name(self) -> str:
        """
        Auxiliary function for determining the cache file name based on the parameters (sigma,
        omega, l, etc.) used for generating the lookup tables.

        Returns:
            Base name of the cache file.
        """
        cache_name = f"cache_score-scaling_s{torch.min(self.sigma_grid).cpu().item():04.3f}-{torch.max(self.sigma_grid).cpu().item():04.3f}-{self.sigma_grid.shape[0]:d}_l{self.l_grid.shape[0]:d}_o{self.num_omega:d}-{self.omega_exponent:d}.npz"
        return cache_name

    def get_sigma_idx(self, sigma: torch.Tensor) -> torch.Tensor:
        """
        Convert continuous sigmas to the indices of the closest tabulated values.

        Args:
            sigma: IGSO3 std devs

        Returns:
            Index tensor mapping the provided sigma values to the internal lookup table.
        """
        return torch.bucketize(sigma, self.sigma_grid)

    def get_score_scaling(self, sigma: torch.Tensor) -> torch.Tensor:
        """
        Get tabulated score scaling values for the closest tabulated noise levels. Since the score
        scaling is pre-computed and tabulated, no gradients will be propagated through it.

        Args:
            sigma: IGSO3 std devs.

        Returns:
            Tensor of score scaling values (same shape as sigma).
        """
        sigma_idx = self.get_sigma_idx(sigma)
        score_scaling = self.score_scaling[sigma_idx]
        return score_scaling.detach()

    def _compute_score_scaling(self, sigma_grid):
        """
        Compute loss scaling factors for the score based on the IGSO(3) probability density and the
        derivative of its logarithm. Variance is scaled by a factor of 1/2 to account for the
        deacceleration of time in the diffusion process due to the choice of SO(3) basis and
        guarantee time-reversibility (see appendix E.3 in [#yim2023_3]_).

        Args:
            sigma_grid: Grid of IGSO3 std devs.

        Returns:
            1D tensor with the same shape as `sigma_grid` containing the score scaling for
            different noise levels.

        References
        ----------
        .. [#yim2023_3] Yim, Trippe, De Bortoli, Mathieu, Doucet, Barzilay, Jaakkola:
           SE(3) diffusion model with application to protein backbone generation.
           arXiv preprint arXiv:2302.02277. 2023.
        """

        current_device = sigma_grid.device
        sigma_grid_tmp = sigma_grid.to(torch.float64)

        # If cuda is available, initialize everything on GPU.
        if torch.cuda.is_available():
            sigma_grid_tmp = sigma_grid_tmp.to(device="cuda")

        # Set up grid for angle resolution. Convert to double precision for better handling of
        # numerics.
        omega_grid = torch.linspace(0.0, 1, self.num_omega).to(sigma_grid_tmp)

        # Increase sample density for lower values
        omega_grid = omega_grid**self.omega_exponent
        omega_grid = omega_grid * np.pi

        # Compute the expansion for all omegas and sigmas.
        pdf_igso3 = generate_igso3_lookup_table(
            omega_grid, sigma_grid_tmp, l_max=self.l_max, tol=self.tol
        )
        # Apply the pre-factor from USO(3).
        prefactor_uniform = uniform_so3_density(omega_grid)
        pdf_igso3 = pdf_igso3 * prefactor_uniform[None, :]
        # Ensure everything is positive.
        pdf_igso3 = torch.abs(pdf_igso3)

        # Compute the derivative of the logarithm of the probability distribution.
        dlog_igso3 = generate_dlog_igso3_lookup_table(
            omega_grid, sigma_grid_tmp, l_max=self.l_max, tol=self.tol
        )

        # Compute scaling of the score for loss as expectation value.
        score_scaling = torch.sqrt(
            torch.sum(dlog_igso3**2 * pdf_igso3, dim=1)
            / (3.0 * torch.sum(pdf_igso3, dim=1) + self.tol)
        )

        score_scaling = score_scaling.to(device=current_device)

        return score_scaling.to(sigma_grid.dtype)

    def forward(self, sigma: torch.Tensor, rotation_vectors: torch.Tensor) -> torch.Tensor:
        """
        Compute the SO(3) score in rotation vector representation.

        Args:
            sigma: Tensor of std devs (with the shape [num_eps]).
            rotation_vectors: Rotation vectors with the shape [num_eps x 3].

        Returns:
            SO(3) score with the shape [num_eps x 3].
        """
        # Compute rotation angles as vector lengths.
        rotation_angles = torch.norm(rotation_vectors, dim=-1)
        # Compute the derivative term coming from the IGSO(3) expansion logarithm.
        dlog_igso3 = dlog_igso3_expansion(rotation_angles, sigma, self.l_grid, tol=self.tol)
        # Compute full score (small offset for numerical stability).
        score = rotation_vectors / (rotation_angles[..., None] + self.tol) * dlog_igso3[..., None]
        return score


def uniform_so3_density(omega: torch.Tensor) -> torch.Tensor:
    """
    Compute the density over the uniform angle distribution in SO(3).

    Args:
        omega: Angles in radians.

    Returns:
        Uniform distribution density.
    """
    return (1.0 - torch.cos(omega)) / np.pi


def igso3_expansion(
    omega: torch.Tensor, sigma: torch.Tensor, l_grid: torch.Tensor, tol=1e-7
) -> torch.Tensor:
    """
    Compute the IGSO(3) angle probability distribution function for pairs of angles and std dev
    levels. The expansion is computed using a grid of expansion orders ranging from 0 to l_max.

    This function approximates the power series in equation 5 of [#yim2023_3]_. With this
    parameterization, IGSO(3) agrees with the Brownian motion on SO(3) with t=sigma^2.

    Args:
        omega: Values of angles (1D tensor).
        sigma: Values of std dev of IGSO3 distribution (1D tensor of same shape as `omega`).
        l_grid: Tensor containing expansion orders (0 to l_max).
        tol: Small offset for numerical stability.

    Returns:
        IGSO(3) angle distribution function (without pre-factor for uniform SO(3) distribution).

    References
    ----------
    .. [#yim2023_3] Yim, Trippe, De Bortoli, Mathieu, Doucet, Barzilay, Jaakkola:
        SE(3) diffusion model with application to protein backbone generation.
        arXiv preprint arXiv:2302.02277. 2023.
    """
    # Pre-compute sine in denominator and clamp for stability.
    denom_sin = torch.sin(0.5 * omega)

    # Pre-compute terms that rely only on expansion orders.
    l_fac_1 = 2.0 * l_grid + 1.0
    l_fac_2 = -l_grid * (l_grid + 1.0)

    # Pre-compute numerator of expansion which only depends on angles.
    numerator_sin = torch.sin((l_grid[None, :] + 1 / 2) * omega[:, None])

    # Pre-compute exponential term with (2l+1) prefactor.
    exponential_term = l_fac_1[None, :] * torch.exp(l_fac_2[None, :] * sigma[:, None] ** 2 / 2)

    # Compute series expansion
    f_igso = torch.sum(exponential_term * numerator_sin, dim=1)
    # For small omega, accumulate limit of sine fraction instead:
    # lim[x->0] sin((l+1/2)x) / sin(x/2) = 2l + 1
    f_limw = torch.sum(exponential_term * l_fac_1[None, :], dim=1)

    # Finalize expansion. Offset for stability can be added since omega is [0,pi] and sin(omega/2)
    # is positive in this interval.
    f_igso = f_igso / (denom_sin + tol)

    # Replace values at small omega with limit.
    f_igso = torch.where(omega <= tol, f_limw, f_igso)

    # Remove remaining numerical problems
    f_igso = torch.where(
        torch.logical_or(torch.isinf(f_igso), torch.isnan(f_igso)), torch.zeros_like(f_igso), f_igso
    )

    return f_igso


def digso3_expansion(
    omega: torch.Tensor, sigma: torch.Tensor, l_grid: torch.Tensor, tol=1e-7
) -> torch.Tensor:
    """
    Compute the derivative of the IGSO(3) angle probability distribution function with respect to
    the angles for pairs of angles and std dev levels. As in `igso3_expansion` a grid is used for the
    expansion levels. Evaluates the derivative directly in order to avoid second derivatives during
    backpropagation.

    The derivative of the angle-dependent part is computed as:

    .. math ::
        \frac{\\partial}{\\partial \\omega} \frac{\\sin((l+\tfrac{1}{2})\\omega)}{\\sin(\tfrac{1}{2}\\omega)} = \frac{l\\sin((l+1)\\omega) - (l+1)\\sin(l\\omega)}{1 - \\cos(\\omega)}

    (obtained via quotient rule + different trigonometric identities).

    Args:
        omega: Values of angles (1D tensor).
        sigma: Values of IGSO3 distribution std devs (1D tensor of same shape as `omega`).
        l_grid: Tensor containing expansion orders (0 to l_max).
        tol: Small offset for numerical stability.

    Returns:
        IGSO(3) angle distribution derivative (without pre-factor for uniform SO(3) distribution).
    """
    denom_cos = 1.0 - torch.cos(omega)

    l_fac_1 = 2.0 * l_grid + 1.0
    l_fac_2 = l_grid + 1.0
    l_fac_3 = -l_grid * l_fac_2

    # Pre-compute numerator of expansion which only depends on angles.
    numerator_sin = l_grid[None, :] * torch.sin(l_fac_2[None, :] * omega[:, None]) - l_fac_2[
        None, :
    ] * torch.sin(l_grid[None, :] * omega[:, None])

    # Compute series expansion
    df_igso = torch.sum(
        l_fac_1[None, :] * torch.exp(l_fac_3[None, :] * sigma[:, None] ** 2 / 2) * numerator_sin,
        dim=1,
    )

    # Finalize expansion. Offset for stability can be added since omega is [0,pi] and cosine term
    # is positive in this interval.
    df_igso = df_igso / (denom_cos + tol)

    # Replace values at small omega with limit (=0).
    df_igso = torch.where(omega <= tol, torch.zeros_like(df_igso), df_igso)

    # Remove remaining numerical problems
    df_igso = torch.where(
        torch.logical_or(torch.isinf(df_igso), torch.isnan(df_igso)),
        torch.zeros_like(df_igso),
        df_igso,
    )

    return df_igso


def dlog_igso3_expansion(
    omega: torch.Tensor, sigma: torch.Tensor, l_grid: torch.Tensor, tol=1e-7
) -> torch.Tensor:
    """
    Compute the derivative of the logarithm of the IGSO(3) angle distribution function for pairs of
    angles and std dev levels:

    .. math ::
        \frac{\\partial}{\\partial \\omega} \\log f(\\omega) = \frac{\tfrac{\\partial}{\\partial \\omega} f(\\omega)}{f(\\omega)}

    Required for SO(3) score computation.

    Args:
        omega: Values of angles (1D tensor).
        sigma: Values of IGSO3 std devs (1D tensor of same shape as `omega`).
        l_grid: Tensor containing expansion orders (0 to l_max).
        tol: Small offset for numerical stability.

    Returns:
        IGSO(3) angle distribution derivative (without pre-factor for uniform SO(3) distribution).
    """
    f_igso3 = igso3_expansion(omega, sigma, l_grid, tol=tol)
    df_igso3 = digso3_expansion(omega, sigma, l_grid, tol=tol)

    return df_igso3 / (f_igso3 + tol)


@torch.no_grad()
def generate_lookup_table(
    base_function: Callable,
    omega_grid: torch.Tensor,
    sigma_grid: torch.Tensor,
    l_max: int = 1000,
    tol: float = 1e-7,
):
    """
    Auxiliary function for generating a lookup table from IGSO(3) expansions and their derivatives.
    Takes a basic function and loops over different std dev levels.

    Args:
        base_function: Function used for setting up the lookup table.
        omega_grid: Grid of angle values ranging from [0,pi] (shape is[num_omega]).
        sigma_grid: Grid of IGSO3 std dev values (shape is [num_sigma]).
        l_max: Number of terms used in the series expansion.
        tol: Small value for numerical stability.

    Returns:
        Table of function values evaluated at different angles and std dev levels. The final shape is
        [num_sigma x num_omega].
    """
    # Generate grid of expansion orders.
    l_grid = torch.arange(l_max + 1, device=omega_grid.device).to(omega_grid.dtype)

    n_omega = len(omega_grid)
    n_sigma = len(sigma_grid)

    # Populate lookup table for different time frames.
    f_table = torch.zeros(n_sigma, n_omega, device=omega_grid.device, dtype=omega_grid.dtype)

    for eps_idx in tqdm(range(n_sigma), desc=f"Computing {base_function.__name__}"):
        f_table[eps_idx, :] = base_function(
            omega_grid,
            torch.ones_like(omega_grid) * sigma_grid[eps_idx],
            l_grid,
            tol=tol,
        )

    return f_table


def generate_igso3_lookup_table(
    omega_grid: torch.Tensor,
    sigma_grid: torch.Tensor,
    l_max: int = 1000,
    tol: float = 1e-7,
) -> torch.Tensor:
    """
    Generate a lookup table for the IGSO(3) probability distribution function of angles.

    Args:
        omega_grid: Grid of angle values ranging from [0,pi] (shape is[num_omega]).
        sigma_grid: Grid of IGSO3 std dev values (shape is [num_sigma]).
        l_max: Number of terms used in the series expansion.
        tol: Small value for numerical stability.

    Returns:
        Table of function values evaluated at different angles and std dev levels. The final shape is
        [num_sigma x num_omega].
    """
    f_igso = generate_lookup_table(
        base_function=igso3_expansion,
        omega_grid=omega_grid,
        sigma_grid=sigma_grid,
        l_max=l_max,
        tol=tol,
    )
    return f_igso


def generate_dlog_igso3_lookup_table(
    omega_grid: torch.Tensor,
    sigma_grid: torch.Tensor,
    l_max: int = 1000,
    tol: float = 1e-7,
) -> torch.Tensor:
    """
    Generate a lookup table for the derivative of the logarithm of the angular IGSO(3) probability
    distribution function. Used e.g. for computing scaling of SO(3) norms.

    Args:
        omega_grid: Grid of angle values ranging from [0,pi] (shape is[num_omega]).
        sigma_grid: Grid of IGSO3 std dev values (shape is [num_sigma]).
        l_max: Number of terms used in the series expansion.
        tol: Small value for numerical stability.

    Returns:
        Table of function values evaluated at different angles and std dev levels. The final shape is
        [num_sigma x num_omega].
    """
    dlog_igso = generate_lookup_table(
        base_function=dlog_igso3_expansion,
        omega_grid=omega_grid,
        sigma_grid=sigma_grid,
        l_max=l_max,
        tol=tol,
    )
    return dlog_igso



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
from pathlib import Path

import pytest

from bioemu.utils import count_samples_in_output_dir, format_npz_samples_filename


@pytest.mark.parametrize("batch_size", [5, 7, 10])
def test_count_samples_in_output_dir(tmp_path: Path, batch_size: int) -> None:
    num_samples = 10
    for seed in range(0, num_samples, batch_size):
        n = min(batch_size, num_samples - seed)
        npz_file_path = tmp_path / format_npz_samples_filename(seed, n)
        npz_file_path.touch()

    assert count_samples_in_output_dir(tmp_path) == num_samples



# To be executed in separate python env. See README.md (section "side-chain reconstruction").
try:
    from hpacker import HPacker

    _HAS_HPACKER = True
except Exception:
    _HAS_HPACKER = False


def _hpacker(protein_pdb_in: str, protein_pdb_out: str) -> None:
    """Call hpacker to reconstruct sidechains.

    Args:
        protein_pdb_in: Input PDB file with backbone only.
        protein_pdb_out: Output PDB file with sidechains reconstructed.

    """
    if not _HAS_HPACKER:
        raise ImportError("hpacker not found, please install hpacker first")
    hpacker = HPacker(protein_pdb_in)
    hpacker.reconstruct_sidechains(num_refinement_iterations=5)
    hpacker.write_pdb(protein_pdb_out)


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument("protein_pdb_in", help="Input PDB file with backbone only.")
    parser.add_argument("protein_pdb_out", help="Output PDB file with sidechains reconstructed.")
    args = parser.parse_args()
    _hpacker(args.protein_pdb_in, args.protein_pdb_out)



import logging
import os
import subprocess

HPACKER_INSTALL_SCRIPT = os.path.join(
    os.path.dirname(os.path.realpath(__file__)), "setup_sidechain_relax.sh"
)
HPACKER_DEFAULT_ENVNAME = "hpacker"
HPACKER_DEFAULT_REPO_DIR = os.path.join(os.path.expanduser("~"), ".hpacker")

logging.basicConfig(level=logging.DEBUG)
logger = logging.getLogger(__name__)


def ensure_hpacker_install(
    envname: str = HPACKER_DEFAULT_ENVNAME, repo_dir: str = HPACKER_DEFAULT_REPO_DIR
) -> None:
    """
    Ensures hpacker and its dependencies are installed under conda environment
    named `envname`
    """
    conda_root = os.getenv("CONDA_ROOT", None)
    if conda_root is None:
        # Attempt $CONDA_PREFIX_1
        conda_root = os.getenv("CONDA_PREFIX_1", None)
    assert conda_root is not None, "conda required to install hpacker environment"
    conda_envs = os.listdir(os.path.join(conda_root, "envs"))

    if envname not in conda_envs:
        logger.info("Setting up hpacker dependencies...")
        _install = subprocess.run(
            ["bash", HPACKER_INSTALL_SCRIPT, envname, repo_dir],
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
        )
        assert (
            _install.returncode == 0
        ), f"Something went wrong during hpacker setup: {_install.stdout.decode()}"



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
import os
import subprocess
from unittest.mock import patch

import numpy as np

from bioemu.get_embeds import StrPath, get_colabfold_embeds, shahexencode

TEST_SEQ = "TESTSEQ"


def mock_run_colabfold(
    input_file: StrPath, res_dir: StrPath, colabfold_env: dict[str, str], msa_host_url: str | None
) -> subprocess.CompletedProcess:
    seq = TEST_SEQ
    seqsha = os.path.basename(input_file).split(".")[0]
    single_rep_tempfile = os.path.join(
        res_dir,
        f"{seqsha}__unknown_description__single_repr_evo_rank_001_alphafold2_model_3_seed_000.npy",
    )
    pair_rep_tempfile = os.path.join(
        res_dir,
        f"{seqsha}__unknown_description__pair_repr_evo_rank_001_alphafold2_model_3_seed_000.npy",
    )
    np.save(single_rep_tempfile, np.random.rand(len(seq), 10))
    np.save(pair_rep_tempfile, np.random.rand(len(seq), len(seq), 10))
    return subprocess.CompletedProcess("mock", returncode=0)


def test_get_colabfold_embeds(tmp_path):
    seq = TEST_SEQ
    cache_embeds_dir = tmp_path / "cache"

    with patch("bioemu.get_embeds.run_colabfold", side_effect=mock_run_colabfold), patch(
        "bioemu.get_embeds.ensure_colabfold_install"
    ):
        result_single, result_pair = get_colabfold_embeds(seq, cache_embeds_dir)

    seqsha = shahexencode(seq)
    single_rep_file = cache_embeds_dir / f"{seqsha}_single.npy"
    pair_rep_file = cache_embeds_dir / f"{seqsha}_pair.npy"

    # Assertions
    assert os.path.exists(single_rep_file)
    assert os.path.exists(pair_rep_file)
    assert result_single == str(single_rep_file)
    assert result_pair == str(pair_rep_file)

    # Now that the files are cached, so we should not run colabfold again
    exist_single, exist_pair = get_colabfold_embeds(seq, cache_embeds_dir)
    assert exist_single == result_single
    assert exist_pair == result_pair



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
import numpy as np
import torch
from scipy.stats import wasserstein_distance

from bioemu.so3_sde import (
    DiGSO3SDE,
    ScoreSO3,
    rotmat_to_rotvec,
    skew_matrix_exponential_map_axis_angle,
    vector_to_skew_matrix,
)


# Exponential map from so(3) to SO(3), this is the matrix exponential
def exp(A):
    rotvec = rotmat_to_rotvec(A)
    angles = torch.norm(rotvec, dim=-1)
    vectors = rotvec / (angles[..., None] + 1e-5)

    # Exponentials using Rodrigues' formula.
    skewmats = -vector_to_skew_matrix(vectors)
    exp_map = skew_matrix_exponential_map_axis_angle(angles, skewmats)
    return exp_map


# Exponential map from tangent space at R0 to SO(3)
def expmap(R0, tangent):
    skew_sym = torch.einsum("...ij,...ik->...jk", R0, tangent)
    return torch.einsum("...ij,...jk->...ik", R0, exp(skew_sym))


# Normal sample in tangent space at R0
def tangent_gaussian(R0):
    return torch.einsum(
        "...ij,...jk->...ik", R0, vector_to_skew_matrix(torch.randn(R0.shape[0], 3))
    )


# Simulates a Geodesic Random Walk (GRW) on a Riemannian manifold.
# Implements Algorithm 1 in https://arxiv.org/pdf/2202.02763.pdf
def geodesic_random_walk(p_initial, drift, diffusion, ts):
    Rts = {ts[0]: p_initial()}
    for i in range(1, len(ts)):
        dt = ts[i] - ts[i - 1]  # negative for reverse process
        Rts[ts[i]] = expmap(
            Rts[ts[i - 1]],
            dt * drift(Rts[ts[i - 1]], ts[i - 1])  # Drift
            + diffusion(ts[i - 1])
            * np.sqrt(abs(dt))
            * tangent_gaussian(Rts[ts[i - 1]]),  # Diffusion
        )
    return Rts


def test_so3_score(
    num_samples: int = 1000,
    t_truncate: float = 0.1,
    dist_threshold: float = 0.2,
):
    """
    Unit test for ensuring forward SDE matches reverse SDE.
    Following https://arxiv.org/pdf/2202.02763.pdf we simulate
    a geodesic random walk (GRW) from algorithm 1 of the forward
    and reverse SDE on manifolds which are given as equations
    (3) and (4) respectively. This test effectively sees if the
    score is able to simulate the reverse process correctly.

    We test if the Wasserstein distance between the distributions of
    every Euler Maruyama discretization step is below a threshold (dist_threshold).

    Args:
        num_samples: Number of rotations to simulate.
            Higher will lead to better distribution estimation but more compute.
        t_truncate: value of t to truncate when comparing distributions.
            Small t can be instable due to the truncated IGSO(3) density and number of samples
        dist_threshold: threshold for the Wasserstein distance of when we consider two distributions
            to be the same.
        timeout: allowed time (in seconds) for the test to run.
    """
    np.random.seed(seed=1)
    torch.manual_seed(1)

    # Set-up SDE
    # We use sigma_max=2.5 to make sure the SDE converges to the uniform density.
    # 1.5 is used in FrameDiff because converging is not necessary.
    sde = DiGSO3SDE(
        eps_t=1e-4,
        sigma_min=0.1,
        sigma_max=2.5,
        l_max=1000,
    )

    # Discretization of [0, T]
    ts = np.linspace(0, sde.T, 200)

    # Simulate forward GRW
    forward_samples = geodesic_random_walk(
        # Surrogate data distribution mocked as identity rotations.
        p_initial=lambda: torch.eye(3)[None].repeat(num_samples, 1, 1),
        drift=lambda Rt, t: 0.0,
        diffusion=lambda t: sde.beta(torch.tensor(t)).type(torch.float64),
        ts=ts,
    )

    igso3_score = ScoreSO3(num_omega=1000, sigma_grid=torch.linspace(0.1, 2.5, 1000))

    def sampling_score_t(Rt, t):
        sigma = sde._marginal_std(torch.tensor(t))
        grad_coefficients = igso3_score.forward(
            sigma * torch.ones(Rt.shape[0]), rotmat_to_rotvec(Rt)
        )
        return torch.einsum("...ij,...jk->...ik", Rt, vector_to_skew_matrix(grad_coefficients))

    # Simulate reverse GRW
    reverse_samples = geodesic_random_walk(
        p_initial=lambda: sde.prior_sampling((num_samples,)),  # Uniform over SO(3).
        drift=lambda Rt, t: -sde.beta(torch.tensor(t)).type(torch.float64) ** 2
        * sampling_score_t(Rt, t),
        diffusion=lambda t: sde.beta(torch.tensor(t)).type(torch.float64),
        ts=ts[::-1],
    )
    for t in ts:
        # Skip calculating distances for t less than a cutoff.
        if t < t_truncate:
            continue
        Rt_forward = rotmat_to_rotvec(forward_samples[t])
        Rt_reverse = rotmat_to_rotvec(reverse_samples[t])

        omega_forward = torch.norm(Rt_forward, dim=-1).numpy()
        omega_reverse = torch.norm(Rt_reverse, dim=-1).numpy()

        dist = wasserstein_distance(omega_forward, omega_reverse)
        assert dist < dist_threshold, f"Wasserstein distance too large for t={t}."



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
"""
Based on code from https://github.com/yang-song/score_sde_pytorch
which is released under Apache licence.

Key change: adapted to work on batched pytorch_geometric style data.
"""

import abc
import logging

import numpy as np
import torch
from torch._prims_common import DeviceLikeType


def _broadcast_like(x, like):
    """
    add broadcast dimensions to x so that it can be broadcast over ``like``
    """
    if like is None:
        return x
    return x[(...,) + (None,) * (like.ndim - x.ndim)]


def maybe_expand(
    x: torch.Tensor, batch: torch.LongTensor | None, like: torch.Tensor = None
) -> torch.Tensor:
    """

    Args:
        x: shape (batch_size, ...)
        batch: shape (num_thingies,) with integer entries in the range [0, batch_size), indicating which sample each thingy belongs to
        like: shape x.shape + potential additional dimensions
    Returns:
        expanded x with shape (num_thingies,), or if given like.shape, containing value of x for each thingy.
        If `batch` is None, just returns `x` unmodified, to avoid pointless work if you have exactly one thingy per sample.
    """
    x = _broadcast_like(x, like)
    if batch is None:
        return x
    else:
        if x.shape[0] == batch.shape[0]:
            logging.warning(
                "Warning: batch shape is == x shape, are you trying to expand something that is already expanded?"
            )
        return x[batch]


class SDE:
    """Corruption using a stochastic differential equation."""

    @abc.abstractmethod
    def sde(
        self, x: torch.Tensor, t: torch.Tensor, batch_idx: torch.LongTensor | None = None
    ) -> tuple[torch.Tensor, torch.Tensor]:
        """Returns drift f and diffusion coefficient g such that dx = f * dt + g * sqrt(dt) * standard Gaussian"""
        pass  # drift: (nodes_per_sample * batch_size, num_features), diffusion (batch_size,)

    @property
    def T(self) -> float:
        return 1.0

    @abc.abstractmethod
    def marginal_prob(
        self,
        x: torch.Tensor,
        t: torch.Tensor,
        batch_idx: torch.LongTensor | None = None,
    ) -> tuple[torch.Tensor, torch.Tensor]:
        """Returns mean and standard deviation of the marginal distribution of the SDE, $p_t(x)$."""
        pass  # mean: (nodes_per_sample * batch_size, num_features), std: (nodes_per_sample * batch_size, 1)

    def mean_coeff_and_std(
        self, x: torch.Tensor, t: torch.Tensor, batch_idx: torch.LongTensor | None = None
    ) -> tuple[torch.Tensor, torch.Tensor]:
        """Returns mean coefficient and standard deviation of marginal distribution at time t."""
        return self.marginal_prob(
            torch.ones_like(x), t, batch_idx
        )  # mean_coeff: same shape as x, std: same shape as x

    def sample_marginal(
        self,
        x: torch.Tensor,
        t: torch.Tensor,
        batch_idx: torch.LongTensor | None = None,
    ) -> torch.Tensor:
        """Sample marginal for x(t) given x(0).
        Returns:
          sampled x(t)
        """
        mean, std = self.marginal_prob(x=x, t=t, batch_idx=batch_idx)
        z = torch.randn_like(x)

        return mean + std * z

    def prior_sampling(
        self,
        shape: torch.Size | tuple,
        device: DeviceLikeType | None = None,
    ) -> torch.Tensor:
        raise NotImplementedError


class BaseVPSDE(SDE):
    """Base class for variance-preserving SDEs of the form
            dx = - 0.5 * beta_t * x * dt + sqrt(beta_t) * z * sqrt(dt)
    where z is unit Gaussian noise, or equivalently
            dx = - 0.5 * beta_t *x * dt + sqrt(beta_t) * dW

    """

    @abc.abstractmethod
    def beta(self, t: torch.Tensor) -> torch.Tensor:
        ...

    @abc.abstractmethod
    def _marginal_mean_coeff(self, t: torch.Tensor) -> torch.Tensor:
        """This should be implemented to compute exp(-0.5 * int_0^t beta(s) ds). See equation (29) of Song et al."""
        ...

    def marginal_prob(
        self,
        x: torch.Tensor,
        t: torch.Tensor,
        batch_idx: torch.LongTensor | None = None,
    ) -> tuple[torch.Tensor, torch.Tensor]:
        mean_coeff = self._marginal_mean_coeff(t)
        mean = maybe_expand(mean_coeff, batch_idx, x) * x
        std = maybe_expand(torch.sqrt(1.0 - mean_coeff**2), batch_idx, x)
        return mean, std

    def sigma(self, t: torch.Tensor):
        return self.marginal_prob(t, t)[1]

    def prior_sampling(
        self,
        shape: torch.Size | tuple,
        device: DeviceLikeType | None = None,
    ) -> torch.Tensor:
        return torch.randn(*shape, device=device)

    def sde(
        self,
        x: torch.Tensor,
        t: torch.Tensor,
        batch_idx: torch.LongTensor | None = None,
    ) -> tuple[torch.Tensor, torch.Tensor]:
        beta_t = self.beta(t)
        drift = -0.5 * maybe_expand(beta_t, batch_idx, x) * x
        diffusion = maybe_expand(torch.sqrt(beta_t), batch_idx, x)
        return drift, diffusion


class CosineVPSDE(BaseVPSDE):
    """Variance-preserving SDE with cosine noise schedule"""

    def __init__(self, s: float = 0.008):
        self.s = s
        self.c = np.cos(s / (1 + s) * np.pi / 2)

    def beta(self, t) -> torch.Tensor:
        # Derived from _marginal_mean_coeff using equation (29) in Song et al.
        return torch.tan((t + self.s) / (1 + self.s) * np.pi / 2) * np.pi / (1 + self.s)

    def _marginal_mean_coeff(self, t: torch.Tensor) -> torch.Tensor:
        mean_coeff = torch.cos((t + self.s) / (1 + self.s) * np.pi / 2) / self.c
        # Horror: torch.cos(np.pi/2) < 0
        return torch.clip(mean_coeff, 0, 1)
